"""Meltano manifest.

A Meltano manifest file is JSON file which representation of all information
required to describe a Meltano project. It conforms to the same schema used for
`meltano.yml`, and can be thought of as a pre-processed Meltano project file,
where most of the processing that would normally be done at run-time has
instead been done when the manifest file was compiled.

Manifest files always have a specific environment denoted in their name (i.e
`meltano-manifest.<Meltano environment name>.json`), or if the name is simply
`meltano-manifest.json` it represents running its Meltano project with no
active environment. As the environment selection is determined by which
manifest file is used, a manifest file will never have an `environments` field.

Like regular Meltano project files, manifest files should not contain secrets.
As such, they cannot be populated with values from environment variables, or
any other external source that may have sensitive data.

Unlike Meltano project files (which have `include_paths`), manifest files are
self-contained. Once a Meltano environment has been selected, there is no need
to reference any other file to fully describe the Meltano project.

Plugin lockfiles are merged into manifest files during compilation time.

Because the manifest file aims to be a single source of truth for project data,
and aims to enable significant simplication of `meltano.core`, few existing
`meltano.core` services are used for the compilation of the manifest file. In
the future, it should be the case that no `meltano.core` serivces are used for
the compilation of the manifest files.

Related: https://github.com/meltano/meltano/issues/7270

Unfortunately not all process can be done during the compilation process.
Things which still need to be processed when using a manifest file as a source
of information include:
- The merger and expansion of terminal env vars into a `.env` file, if one
  exists. Let the result of this merger be called env vars "A".
- The merger and expansion of env vars "A" into the job env, if a job is
  being run. Let the result of this merger be called env vars "B".
- The merger and expansion of env vars "B" into the schedule env, if a
  schedule is being run. Let the result of this merger be called env vars "C".
- The merger and expansion of env vars "C" into all other string fields of the
  manifest file (with exceptions) aside from the `plugins` field. This is the
  project-level env vars.
- The merger and expansion of the project-level env vars into a plugin env, if
  a plugin is being run. This is the plugin-level env vars.

Recognizing that this process is strictly hierarchical, the manifest module
provides a contexts submodule, which has context manager functions. These can
be used to establish an active manifest file, and then perform the necessary
processing for any call to get settings or env vars from the manifest within
the established context.

Under this paradigm, a function within `meltano.core` need not be provided the
job, schedule, plugin, etc. that it's working with. Instead the appropriate
contexts can be set by the caller, and the lower-level function can get values
of interest by requesting it from within the context. How exactly this will be
implemented is yet to be determined.

Related: https://github.com/meltano/meltano/issues/7271
"""

from __future__ import annotations

from meltano.core.manifest.manifest import Manifest



from __future__ import annotations  # noqa: D100

from abc import ABC, abstractmethod


class IncompatibleVersionError(Exception):
    """A component is incompatible with its representation."""

    def __init__(self, message, file_version: int, version: int):  # noqa: ANN001, D107
        super().__init__(message)

        self.file_version = file_version
        self.version = version


class Versioned(ABC):
    """Mixin to represent something that must be compatible with a certain version."""

    @property
    @abstractmethod
    def file_version(self) -> int:  # noqa: D102
        pass

    def is_compatible(self, version: int | None = None) -> bool | None:  # noqa: D102
        try:
            self.ensure_compatible(version=version)
            return True
        except IncompatibleVersionError:
            return False

    def ensure_compatible(self, version: int | None = None) -> None:  # noqa: D102
        version = self.__class__.__version__ if version is None else version
        file_version = self.file_version

        if file_version != version:
            raise IncompatibleVersionError(
                f"Version {version} required, currently at {self.file_version}",  # noqa: EM102
                file_version,
                version,
            )



from __future__ import annotations

import pytest

from meltano.core.setting_definition import SettingDefinition, SettingKind


class TestSettingDefinition:
    @pytest.mark.parametrize(
        ("setting_definition", "uncast_expected_pairs"),
        (
            (
                SettingDefinition(name="test_setting", kind=SettingKind.ARRAY),
                (
                    ('["abc", "xyz"]', ["abc", "xyz"]),
                    ('["abc", "xyz",]', ["abc", "xyz"]),
                    ("['abc', 'xyz']", ["abc", "xyz"]),
                    ("[abc, xyz]", ValueError),
                    ("'abc', 'xyz'", ["abc", "xyz"]),
                    ("'abc', 'xyz',", ["abc", "xyz"]),
                    ('[null, "xyz"]', [None, "xyz"]),
                    ("{'abc', 'xyz'}", ValueError),
                    ("1234", ValueError),
                    ("[1234, 5678]", [1234, 5678]),
                ),
            ),
            (
                SettingDefinition(name="test_setting", kind=SettingKind.OBJECT),
                (
                    (
                        '{"key_1": "value_1", "key_2": "value_2"}',
                        {"key_1": "value_1", "key_2": "value_2"},
                    ),
                    ('{"key": "value",}', {"key": "value"}),
                    ('{"key": null}', {"key": None}),
                    ("{'key': null}", ValueError),
                    ("key: value", ValueError),
                    ("{'key': 'value'}", {"key": "value"}),
                    ("{key: value}", ValueError),
                    ("{'abc', 'xyz'}", ValueError),
                    ("1234", ValueError),
                ),
            ),
        ),
        ids=("array", "object"),
    )
    def test_cast_value_array(
        self,
        setting_definition: SettingDefinition,
        uncast_expected_pairs,
    ) -> None:
        for uncast, expected in uncast_expected_pairs:
            if isinstance(expected, type) and issubclass(expected, Exception):
                with pytest.raises(expected):
                    setting_definition.cast_value(uncast)
            else:
                assert setting_definition.cast_value(uncast) == expected

    def test_cast_options(self) -> None:
        setting_definition = SettingDefinition(
            name="test_setting",
            kind=SettingKind.OPTIONS,
            options=[
                {"value": "abc", "label": "ABC"},
                {"value": "xyz", "label": "XYZ"},
            ],
        )

        assert setting_definition.cast_value("abc") == "abc"
        assert setting_definition.cast_value("xyz") == "xyz"
        assert setting_definition.cast_value(None) is None
        with pytest.raises(ValueError, match="is not a valid choice"):
            setting_definition.cast_value("def")

    @pytest.mark.parametrize(
        ("setting_definition", "sensitive", "kind"),
        (
            pytest.param(
                SettingDefinition(
                    name="test_setting",
                    sensitive=True,
                    kind=SettingKind.STRING,
                ),
                True,
                SettingKind.STRING,
                id="sensitive-string",
            ),
            pytest.param(
                SettingDefinition(
                    name="test_setting",
                    sensitive=True,
                    kind=SettingKind.PASSWORD,
                ),
                True,
                SettingKind.STRING,
                id="sensitive-password",
            ),
            pytest.param(
                SettingDefinition(
                    name="test_setting",
                    sensitive=False,
                    kind=None,
                ),
                False,
                None,
                id="sensitive-no-kind",
            ),
            pytest.param(
                SettingDefinition(
                    name="test_setting",
                    sensitive=False,
                    kind=SettingKind.STRING,
                ),
                False,
                SettingKind.STRING,
                id="non-sensitive-string",
            ),
            pytest.param(
                SettingDefinition(
                    name="test_setting",
                    sensitive=False,
                    kind=SettingKind.PASSWORD,
                ),
                True,
                SettingKind.PASSWORD,
                id="non-sensitive-password",
            ),
            pytest.param(
                SettingDefinition(
                    name="test_setting",
                    sensitive=False,
                    kind=None,
                ),
                False,
                None,
                id="non-sensitive-no-kind",
            ),
            pytest.param(
                SettingDefinition(
                    name="test_setting",
                    sensitive=None,
                    kind=SettingKind.STRING,
                ),
                None,
                SettingKind.STRING,
                id="no-sensitive-string",
            ),
            pytest.param(
                SettingDefinition(
                    name="test_setting",
                    sensitive=None,
                    kind=SettingKind.PASSWORD,
                ),
                True,
                SettingKind.PASSWORD,
                id="no-sensitive-password",
            ),
            pytest.param(
                SettingDefinition(
                    name="test_setting",
                    sensitive=None,
                    kind=None,
                ),
                None,
                None,
                id="no-sensitive-no-kind",
            ),
        ),
    )
    def test_parse(
        self,
        setting_definition: SettingDefinition,
        *,
        sensitive: bool,
        kind: SettingKind,
    ) -> None:
        assert setting_definition.sensitive is sensitive
        assert setting_definition.kind is kind



from __future__ import annotations

import contextlib
import os

import pytest
import sqlalchemy
from sqlalchemy import create_engine, text
from sqlalchemy.pool import NullPool


def recreate_database(engine, db_name) -> None:
    """Drop & Create a new database, PostgreSQL only."""
    with contextlib.suppress(sqlalchemy.exc.ProgrammingError), engine.begin() as conn:
        conn.execute(text(f"DROP DATABASE {db_name}"))

    with contextlib.suppress(sqlalchemy.exc.ProgrammingError), engine.begin() as conn:
        conn.execute(text(f"CREATE DATABASE {db_name}"))


@pytest.fixture(scope="session")
def engine_uri(worker_id: str) -> str:
    host = os.getenv("POSTGRES_ADDRESS")
    port = os.getenv("POSTGRES_PORT", 5432)
    user = os.getenv("POSTGRES_USER")
    password = os.getenv("POSTGRES_PASSWORD")
    database = f"{os.getenv('POSTGRES_DB', 'pytest_meltano')}_{worker_id}"

    # create the database
    engine_uri = f"postgresql://{user}:{password}@{host}:{port}/postgres"
    engine = create_engine(
        engine_uri,
        isolation_level="AUTOCOMMIT",
        poolclass=NullPool,
        future=True,
    )
    recreate_database(engine, database)

    return f"postgresql://{user}:{password}@{host}:{port}/{database}"



from __future__ import annotations

import random
import typing as t

from meltano.core.tracking.contexts.environment import EnvironmentContext
from meltano.core.utils import hash_sha256

if t.TYPE_CHECKING:
    import pytest


def test_notable_flag_env_vars(monkeypatch: pytest.MonkeyPatch) -> None:
    for notable_flag_env_var in EnvironmentContext.notable_flag_env_vars:
        monkeypatch.delenv(notable_flag_env_var, raising=False)

    assert EnvironmentContext().data["notable_flag_env_vars"] == {}

    def check(env_var_values: tuple[str, ...], expected: bool | None) -> None:
        for notable_flag_env_var in EnvironmentContext.notable_flag_env_vars:
            monkeypatch.setenv(
                notable_flag_env_var,
                random.choice(env_var_values),  # noqa: S311
            )

        assert EnvironmentContext().data["notable_flag_env_vars"] == dict.fromkeys(
            EnvironmentContext.notable_flag_env_vars,
            expected,
        )

    check(("Yes", "TRUE", "1", "true", "y", "on", "t"), expected=True)
    check(("No", "FALSE", "0", "false", "n", "off", "f"), expected=False)
    check(("trew", "Fallse", "nah", "si", "okay", "01"), expected=None)


def test_notable_hashed_env_vars(monkeypatch: pytest.MonkeyPatch) -> None:
    for notable_hashed_env_var in EnvironmentContext.notable_hashed_env_vars:
        monkeypatch.delenv(notable_hashed_env_var, raising=False)
    assert EnvironmentContext().data["notable_hashed_env_vars"] == {}

    github_username = "test_user"
    monkeypatch.setenv("GITHUB_USER", github_username)
    assert EnvironmentContext().data["notable_hashed_env_vars"] == {
        "GITHUB_USER": hash_sha256(github_username),
    }

    github_repo = "test_owner/test_repo"
    monkeypatch.setenv("GITHUB_REPOSITORY", github_repo)
    assert EnvironmentContext().data["notable_hashed_env_vars"] == {
        "GITHUB_USER": hash_sha256(github_username),
        "GITHUB_REPOSITORY": hash_sha256(github_repo),
    }



"""Meltano runtime environments."""

from __future__ import annotations

import copy
import typing as t

from meltano.core.behavior import NameEq
from meltano.core.behavior.canonical import Canonical
from meltano.core.job_state import STATE_ID_COMPONENT_DELIMITER
from meltano.core.plugin import PluginType
from meltano.core.plugin.base import PluginRef
from meltano.core.setting_definition import SettingDefinition
from meltano.core.utils import NotFound

TEnv = t.TypeVar("TEnv", bound="Environment")


class NoActiveEnvironment(Exception):
    """Exception raised when invocation has no active environment."""

    def __init__(self) -> None:
        """Create a new exception."""
        super().__init__("No active environment found")


class EnvironmentNameContainsStateIdDelimiterError(Exception):
    """An environment name contains the state ID component delimiter."""

    def __init__(self, name: str):
        """Create a new exception.

        Args:
            name: The name of the environment.
        """
        super().__init__(
            f"The environment name '{name}' cannot contain the state ID component "
            f"delimiter string '{STATE_ID_COMPONENT_DELIMITER}'",
        )


class EnvironmentPluginConfig(PluginRef):
    """Environment configuration for a plugin."""

    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        config: dict | None = None,
        env: dict | None = None,
        **extras,  # noqa: ANN003
    ):
        """Create a new plugin configuration object.

        Args:
            plugin_type: Extractor, loader, etc.
            name: Name of the plugin.
            config: Plugin configuration.
            env: Plugin environment variables.
            extras: Plugin extras.
        """
        super().__init__(plugin_type, name)
        self.config = copy.deepcopy(config or {})
        self.env = copy.deepcopy(env or {})
        self.extras = extras

    @property
    def extra_config(self):  # noqa: ANN201
        """Get extra config from the Meltano environment, like `select` and `schema`.

        Returns:
            Extra config.
        """
        return {f"_{key}": value for key, value in self.extras.items()}

    @property
    def config_with_extras(self):  # noqa: ANN201
        """Get plugin configuration values from the Meltano environment.

        Returns:
            Plugin configuration with extra values.
        """
        return {**self.config, **self.extra_config}

    @config_with_extras.setter
    def config_with_extras(self, new_config_with_extras: dict[str, t.Any]) -> None:
        """Set plugin configuration values from the Meltano environment.

        Args:
            new_config_with_extras: New plugin configuration with extra values.
        """
        self.config.clear()
        self.extras.clear()

        for key, value in new_config_with_extras.items():
            if key.startswith("_"):
                self.extras[key[1:]] = value
            else:
                self.config[key] = value

    def get_orphan_settings(
        self,
        existing: t.Iterable[SettingDefinition],
    ) -> list[SettingDefinition]:
        """Get orphan settings for this plugin.

        Orphan settings are `config` entries that do not have a
        matching parent entry within `settings`.

        Args:
            existing: Existing settings.

        Returns:
            List of orphan settings.
        """
        return SettingDefinition.from_missing(existing, self.config_with_extras)


class EnvironmentConfig(Canonical):
    """Meltano environment configuration."""

    def __init__(self, plugins: dict[str, list[dict]] | None = None, **extras):  # noqa: ANN003
        """Create a new environment configuration.

        Args:
            plugins: Mapping of plugin types to arrays of plugin configurations.
            extras: Environment extras.
        """
        super().__init__(extras=extras)
        self.plugins = self.load_plugins(plugins or {})

    def load_plugins(
        self,
        plugins: dict[str, list[dict]],
    ) -> dict[PluginType, list[EnvironmentPluginConfig]]:
        """Create plugin configurations from raw dictionary.

        Args:
            plugins: Plugin configurations.

        Returns:
            A mapping of plugin type to arrays of plugin configurations.
        """
        plugin_mapping = {}
        for raw_plugin_type, raw_plugins in plugins.items():
            plugin_type = PluginType(raw_plugin_type)
            plugin_mapping[plugin_type] = [
                EnvironmentPluginConfig(plugin_type=plugin_type, **raw_plugin)
                for raw_plugin in raw_plugins
            ]

        return plugin_mapping


class Environment(NameEq, Canonical):
    """Runtime environment for Meltano runs."""

    def __init__(
        self,
        name: str,
        config: dict | None = None,
        env: dict | None = None,
        state_id_suffix: str | None = None,
    ) -> None:
        """Create a new environment object.

        Args:
            name: Environment name. Must be unique.
            config: Dictionary with environment configuration.
            env: Optional override environment values.
            state_id_suffix: State ID suffix to use.

        Raises:
            EnvironmentNameContainsStateIdDelimiterError: If the name contains the state
                ID component delimiter string.
        """
        if STATE_ID_COMPONENT_DELIMITER in name:
            raise EnvironmentNameContainsStateIdDelimiterError(name)

        super().__init__()

        self.name = name
        self.config = EnvironmentConfig(**(config or {}))
        self.env = env or {}
        self.state_id_suffix = state_id_suffix

    @classmethod
    def find(cls: type[TEnv], objects: t.Iterable[TEnv], name: str) -> TEnv:
        """Lookup an environment by name from an iterable.

        Args:
            objects: Iterable of objects to search.
            name: Environment name.

        Returns:
            Environment object if found.

        Raises:
            NotFound: If an environment is not found.
        """
        try:
            return next(
                env if isinstance(env, cls) else cls.parse(env)
                for env in objects
                if getattr(env, "name", env["name"]) == name
            )
        except StopIteration as stop:
            raise NotFound(name, cls) from stop

    def get_plugin_config(
        self,
        plugin_type: PluginType,
        name: str,
    ) -> EnvironmentPluginConfig:
        """Get configuration for a plugin in this environment.

        Args:
            plugin_type: Extractor, loader, etc.
            name: Plugin name.

        Returns:
            A plugin configuration object if one is present in this environment.
        """
        default = EnvironmentPluginConfig(plugin_type, name)

        return next(
            filter(
                lambda plugin: plugin.name == name,
                self.config.plugins.get(plugin_type, []),
            ),
            default,
        )



from __future__ import annotations

import typing as t

import mock
import pytest

from meltano.core.job_state import STATE_ID_COMPONENT_DELIMITER
from meltano.core.plugin import PluginType, Variant
from meltano.core.plugin.base import PluginRefNameContainsStateIdDelimiterError
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin.singer import SingerTap
from meltano.core.project_plugins_service import PluginDefinitionNotFoundError

if t.TYPE_CHECKING:
    from collections import Counter

    from meltano.core.project import Project
    from meltano.core.project_add_service import ProjectAddService


class TestProjectAddService:
    @pytest.fixture()
    def subject(self, project_add_service):
        return project_add_service

    def test_missing_plugin_exception(self, subject, hub_request_counter) -> None:
        with pytest.raises(PluginDefinitionNotFoundError):
            subject.add(PluginType.EXTRACTORS, "tap-missing")

        assert hub_request_counter["/extractors/index"] == 1
        assert len(hub_request_counter) == 1

    @pytest.mark.order(0)
    @pytest.mark.parametrize(
        ("plugin_type", "plugin_name", "variant", "default_variant"),
        (
            (PluginType.EXTRACTORS, "tap-mock", "meltano", "meltano"),
            (PluginType.LOADERS, "target-mock", None, "original"),
            (PluginType.TRANSFORMERS, "transformer-mock", None, "original"),
            (PluginType.TRANSFORMS, "tap-mock-transform", None, "original"),
            (PluginType.UTILITIES, "utility-mock", None, "original"),
        ),
    )
    def test_add(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant: str,
        default_variant: str,
        subject: ProjectAddService,
        project: Project,
        hub_request_counter: Counter,
    ) -> None:
        used_variant = variant or default_variant
        lockfile_path = project.plugin_lock_path(
            plugin_type,
            plugin_name,
            variant_name=used_variant,
        )
        assert not lockfile_path.exists()

        added = subject.add(plugin_type, plugin_name)
        assert added in project.meltano["plugins"][plugin_type]
        assert lockfile_path.exists()

        # Variant and pip_url are repeated in
        # canonical representation for `meltano.yml`
        canonical = added.canonical()

        if default_variant:
            assert added.variant == default_variant
            assert canonical["variant"] == default_variant
        else:
            assert added.variant == Variant.DEFAULT_NAME
            assert "variant" not in canonical

        assert canonical["pip_url"] == added.pip_url

        assert hub_request_counter[f"/{plugin_type}/index"] == 1
        assert hub_request_counter[f"/{plugin_type}/{plugin_name}--{used_variant}"] == 1
        assert len(hub_request_counter) == 2

    def test_add_inherited(
        self,
        tap,
        subject,
        project: Project,
        hub_request_counter,
    ) -> None:
        # Make sure tap-mock is not in the project as a project plugin
        project.plugins.remove_from_file(tap)

        # Inheriting from base plugin
        inherited = subject.add(
            PluginType.EXTRACTORS,
            "tap-mock-inherited",
            inherit_from="tap-mock",
        )
        assert inherited.canonical() == {
            "name": "tap-mock-inherited",
            "inherit_from": "tap-mock",
            "variant": "meltano",
            "pip_url": "tap-mock",
        }

        # Inheriting from other project plugin
        inception = subject.add(
            PluginType.EXTRACTORS,
            "tap-mock-inception",
            inherit_from="tap-mock-inherited",
        )
        assert inception.canonical() == {
            "name": "tap-mock-inception",
            "inherit_from": "tap-mock-inherited",
        }

        assert hub_request_counter["/extractors/index"] == 1
        assert hub_request_counter["/extractors/tap-mock--meltano"] == 1
        assert len(hub_request_counter) == 2

    @pytest.mark.order(after="test_add_inherited")
    def test_lockfile_inherited(
        self,
        subject: ProjectAddService,
        hub_request_counter: Counter,
    ) -> None:
        child = subject.add(
            PluginType.EXTRACTORS,
            "tap-mock-inherited-new",
            inherit_from="tap-mock",
        )
        assert isinstance(child.parent, SingerTap)
        assert child.parent.name == "tap-mock"

        parent_path = subject.project.plugin_lock_path(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant_name="meltano",
        )
        assert parent_path.stem == "tap-mock--meltano"
        assert parent_path.exists()

        child_path = subject.project.plugin_lock_path(
            child.type,
            child.name,
            variant_name=child.variant,
        )
        assert child_path.stem == "tap-mock-inherited-new--meltano"
        assert not child_path.exists()

        grandchild = subject.add(
            PluginType.EXTRACTORS,
            "tap-mock-inherited-new-2",
            inherit_from="tap-mock-inherited-new",
        )
        assert isinstance(grandchild.parent, ProjectPlugin)
        assert grandchild.parent.name == "tap-mock-inherited-new"
        assert grandchild.parent.parent.name == "tap-mock"

        grandchild_path = subject.project.plugin_lock_path(
            grandchild.type,
            grandchild.name,
            variant_name=grandchild.variant,
        )
        assert grandchild_path.stem == "tap-mock-inherited-new-2--meltano"
        assert not grandchild_path.exists()

        matches = list(parent_path.parent.glob("tap-mock-inherited-new*"))
        assert not matches

        assert hub_request_counter["/extractors/index"] == 1
        assert hub_request_counter["/extractors/tap-mock--meltano"] == 1
        assert len(hub_request_counter) == 2

    def test_add_name_contains_state_id_component_delimiter(
        self,
        subject: ProjectAddService,
    ) -> None:
        with pytest.raises(PluginRefNameContainsStateIdDelimiterError):
            subject.add(
                PluginType.EXTRACTORS,
                f"tap-mock{STATE_ID_COMPONENT_DELIMITER}",
            )

    @mock.patch("meltano.core.plugin_lock_service.PluginLock.save")
    def test_add_update(
        self,
        lock_save: mock.MagicMock,
        target,
        subject: ProjectAddService,
        project: Project,
        hub_request_counter: Counter,
    ) -> None:
        target.config = {
            "username": "meltano",
            "password": "meltano",
        }

        project.plugins.update_plugin(target)

        updated_attrs = {
            "label": "Mock",
            "description": "Mock target",
            "executable": "./target-mock.sh",
            "settings": [
                {"name": "username"},
                {"name": "password"},
            ],
        }

        assert not target.canonical().items() >= updated_attrs.items()

        updated = subject.add(
            target.type,
            target.name,
            update=True,
            **updated_attrs,
        )

        assert hub_request_counter["/loaders/index"] == 1
        assert hub_request_counter["/loaders/target-mock--original"] == 1
        assert len(hub_request_counter) == 2

        assert lock_save.call_count == 1

        assert updated in project.meltano["plugins"][target.type]
        assert updated.canonical().items() >= updated_attrs.items()
        assert updated.config_with_extras

    @mock.patch("meltano.core.plugin_lock_service.PluginLock.save")
    def test_add_update_custom(
        self,
        lock_save: mock.MagicMock,
        subject: ProjectAddService,
        project: Project,
        hub_request_counter: Counter,
    ) -> None:
        custom_plugin = subject.add(
            PluginType.EXTRACTORS,
            "tap-custom",
            namespace="tap_custom",
            config={
                "username": "meltano",
                "password": "meltano",
                "start_date": "2023-01-01",
            },
        )

        updated_attrs = {
            "label": "Custom",
            "description": "Custom tap",
            "executable": "./tap-custom.sh",
            "settings": [
                {"name": "username"},
                {"name": "password"},
            ],
        }

        updated = subject.add(
            custom_plugin.type,
            custom_plugin.name,
            update=True,
            namespace=custom_plugin.namespace,
            **updated_attrs,
        )

        assert len(hub_request_counter) == 0

        assert lock_save.call_count == 0

        assert updated in project.meltano["plugins"][custom_plugin.type]
        assert updated.canonical().items() >= updated_attrs.items()
        assert updated.config_with_extras



from __future__ import annotations

from contextlib import contextmanager

import mock
import pytest

from meltano.core.environment import Environment
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.setting_definition import SettingDefinition
from meltano.core.settings_service import SettingsService
from meltano.core.settings_store import (
    AutoStoreManager,
    DotEnvStoreManager,
    EnvStoreManager,
    InheritedStoreManager,
    MeltanoEnvStoreManager,
    MeltanoYmlStoreManager,
    SettingsStoreManager,
    SettingValueStore,
    StoreNotSupportedError,
)

Store = SettingValueStore


class DummySettingsService(SettingsService):
    """Dummy SettingsService for testing."""

    def __init__(self, *args, **kwargs) -> None:
        """Instantiate new DummySettingsService instance.

        Args:
            args: Positional arguments to pass to the superclass.
            kwargs: Keyword arguments to pass to the superclass.
        """
        super().__init__(*args, **kwargs)

        self.__meltano_yml_config = {}
        self._meltano_environment_config = {}
        self.__definitions = [
            SettingDefinition(name="regular", aliases=["basic"], value="from_default"),
            SettingDefinition(name="password", sensitive=True),
            SettingDefinition(name="env_specific", env_specific=True),
        ]
        self._inherited_settings = None

    @property
    def label(self) -> str:
        return "Dummy"

    @property
    def project_settings_service(self):
        return ProjectSettingsService(Project.find())

    @property
    def docs_url(self) -> str:
        return "https://docs.meltano.com/"

    @property
    def env_prefixes(self):
        return ["dummy"]

    @property
    def db_namespace(self) -> str:
        return "dummy"

    @property
    def setting_definitions(self):
        return self.__definitions

    @property
    def meltano_yml_config(self):
        return self.__meltano_yml_config

    def update_meltano_yml_config(self, config) -> None:
        self.__meltano_yml_config = config

    def update_meltano_environment_config(self, config) -> None:
        self._meltano_environment_config = config

    @property
    def inherited_settings_service(self):
        return self._inherited_settings

    def process_config(self, config):
        return config

    @property
    def environment_config(self) -> dict:
        return self._meltano_environment_config


@pytest.fixture()
def dummy_settings_service(project):
    return DummySettingsService(project)


@pytest.fixture()
def unsupported():
    @contextmanager
    def _unsupported(store):
        with mock.patch.object(
            store.manager,
            "ensure_supported",
            side_effect=StoreNotSupportedError,
        ), mock.patch.object(
            store.manager,
            "set",
            side_effect=StoreNotSupportedError,
        ), mock.patch.object(
            store.manager,
            "unset",
            side_effect=StoreNotSupportedError,
        ), mock.patch.object(
            store.manager,
            "reset",
            side_effect=StoreNotSupportedError,
        ):
            yield

    return _unsupported


class TestAutoStoreManager:
    @pytest.fixture()
    def subject(self, dummy_settings_service, session):
        manager = AutoStoreManager(dummy_settings_service, session=session, cache=False)
        yield manager
        manager.reset()

    @pytest.fixture()
    def set_value_store(self, subject):
        def _set_value_store(value, store, name="regular") -> None:
            subject.manager_for(store).set(
                name,
                [name],
                value,
                setting_def=subject.find_setting(name),
            )

        return _set_value_store

    @pytest.fixture()
    def environment(self):
        return Environment("testing", {})

    @pytest.fixture()
    def assert_value_source(self, subject):
        def _assert_value_source(value, source, name="regular") -> None:
            new_value, metadata = subject.get(
                name,
                setting_def=subject.find_setting(name),
            )
            assert new_value == value
            assert metadata["source"] == source

        return _assert_value_source

    @pytest.mark.parametrize(
        ("setting_name", "preferred_store"),
        (
            ("unknown", Store.MELTANO_YML),
            ("regular", Store.MELTANO_YML),
            ("password", Store.DOTENV),
            ("env_specific", Store.DOTENV),
        ),
    )
    def test_auto_store(
        self,
        setting_name,
        preferred_store,
        subject,
        project,
        unsupported,
        environment,
        monkeypatch,
    ) -> None:
        assert subject.auto_store(setting_name) == preferred_store

        # Meltano environment is selected only when there's an active environment
        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            if preferred_store == Store.MELTANO_YML:
                assert subject.auto_store(setting_name) == Store.MELTANO_ENV

        with unsupported(Store.DOTENV):
            # Sensitive settings won't fall back on `meltano.yml`
            if setting_name == "password":
                assert subject.auto_store(setting_name) == Store.DB
            else:
                assert subject.auto_store(setting_name) == Store.MELTANO_YML
                with monkeypatch.context() as mpc:
                    mpc.setattr(project, "environment", environment)
                    assert subject.auto_store(setting_name) == Store.MELTANO_ENV

        with unsupported(Store.MELTANO_YML):
            # fall back on dotenv
            if preferred_store == Store.MELTANO_YML:
                assert subject.auto_store(setting_name) == Store.DOTENV
                with monkeypatch.context() as mpc:
                    mpc.setattr(project, "environment", environment)
                    assert subject.auto_store(setting_name) == Store.DOTENV

        with unsupported(Store.DOTENV), unsupported(Store.MELTANO_YML):
            assert subject.auto_store(setting_name) == Store.DB
            with monkeypatch.context() as mpc:
                mpc.setattr(project, "environment", environment)
                assert subject.auto_store(setting_name) == Store.DB

            with unsupported(Store.DB):
                assert subject.auto_store(setting_name) is None

        monkeypatch.setattr(project, "readonly", True)

        assert subject.auto_store(setting_name) == Store.DB

    @pytest.mark.usefixtures("assert_value_source")
    def test_get(
        self,
        subject,
        project,
        dummy_settings_service,
        set_value_store,
        monkeypatch,
        environment,
    ) -> None:
        value, metadata = subject.get("regular")
        assert value == "from_default"
        assert metadata["source"] == Store.DEFAULT
        assert metadata["auto_store"] == Store.MELTANO_YML
        assert metadata["overwritable"] is True

        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            value, metadata = subject.get("regular")
            assert value == "from_default"
            assert metadata["source"] == Store.DEFAULT
            assert metadata["auto_store"] == Store.MELTANO_ENV
            assert metadata["overwritable"] is True

        set_value_store("from_db", Store.DB)
        value, metadata = subject.get("regular")
        assert value == "from_db"
        assert metadata["source"] == Store.DB
        assert metadata["auto_store"] == Store.MELTANO_YML
        assert metadata["overwritable"] is True

        set_value_store("from_meltano_yml", Store.MELTANO_YML)
        value, metadata = subject.get("regular")
        assert value == "from_meltano_yml"
        assert metadata["source"] == Store.MELTANO_YML
        assert metadata["auto_store"] == Store.MELTANO_YML
        assert metadata["overwritable"] is True

        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            set_value_store("from_meltano_env", Store.MELTANO_ENV)
            value, metadata = subject.get("regular")
            assert value == "from_meltano_env"
            assert metadata["source"] == Store.MELTANO_ENV
            assert metadata["auto_store"] == Store.MELTANO_ENV
            assert metadata["overwritable"] is True

        set_value_store("from_dotenv", Store.DOTENV)
        value, metadata = subject.get("regular")
        assert value == "from_dotenv"
        assert metadata["source"] == Store.DOTENV
        assert metadata["auto_store"] == Store.MELTANO_YML
        assert metadata["overwritable"] is False

        setting_def = subject.find_setting("regular")
        monkeypatch.setenv(dummy_settings_service.setting_env(setting_def), "from_env")
        value, metadata = subject.get("regular")
        assert value == "from_env"
        assert metadata["source"] == Store.ENV
        assert metadata["auto_store"] == Store.MELTANO_YML
        assert metadata["overwritable"] is False

        monkeypatch.setitem(
            dummy_settings_service.config_override,
            "regular",
            "from_config_override",
        )
        value, metadata = subject.get("regular")
        assert value == "from_config_override"
        assert metadata["source"] == Store.CONFIG_OVERRIDE
        assert metadata["auto_store"] == Store.MELTANO_YML
        assert metadata["overwritable"] is False

        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            value, metadata = subject.get("regular")
            assert value == "from_config_override"
            assert metadata["source"] == Store.CONFIG_OVERRIDE
            assert metadata["auto_store"] == Store.MELTANO_ENV
            assert metadata["overwritable"] is False

        monkeypatch.setattr(project, "readonly", True)
        value, metadata = subject.get("regular")
        assert value == "from_config_override"
        assert metadata["source"] == Store.CONFIG_OVERRIDE
        assert metadata["auto_store"] == Store.DB
        assert metadata["overwritable"] is False

    @pytest.mark.usefixtures("set_value_store")
    def test_set(
        self,
        subject: SettingsStoreManager,
        project,
        unsupported,
        assert_value_source,
        monkeypatch,
        environment,
    ) -> None:
        def set_value(value):
            return subject.set("regular", ["regular"], value)

        # Allow setting to a default value
        assert_value_source("from_default", Store.DEFAULT)
        metadata = set_value("from_default")
        assert metadata["store"] == Store.MELTANO_YML
        assert_value_source("from_default", Store.MELTANO_YML)

        # Falls back on `meltano.yml` when current source is not writable
        with unsupported(Store.DB):
            metadata = set_value("from_meltano_yml")
            assert metadata["store"] == Store.MELTANO_YML
            assert_value_source("from_meltano_yml", Store.MELTANO_YML)

        # Stores in Meltano Environment if active
        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            metadata = set_value("from_meltano_env")
            assert metadata["store"] == Store.MELTANO_ENV
            assert_value_source("from_meltano_env", Store.MELTANO_ENV)

        # Stores in `meltano.yml` by default
        metadata = set_value("from_meltano_yml")
        assert metadata["store"] == Store.MELTANO_YML
        assert_value_source("from_meltano_yml", Store.MELTANO_YML)

        with unsupported(Store.MELTANO_YML):
            # Falls back on `.env` when current store is not supported
            metadata = set_value("from_dotenv")
            assert metadata["store"] == Store.DOTENV
            assert_value_source("from_dotenv", Store.DOTENV)

        # Falls back on `meltano.yml` when `.env` is not supported
        with unsupported(Store.DOTENV):
            metadata = set_value("from_meltano_yml")
            assert metadata["store"] == Store.MELTANO_YML
            # Even though `.env` can't be overwritten
            assert_value_source("from_dotenv", Store.DOTENV)

        # Falls back on system database when
        # neither `.env` or `meltano.yml` are supported
        with unsupported(Store.DOTENV), unsupported(Store.MELTANO_YML):
            metadata = set_value("from_db")
            assert metadata["store"] == Store.DB
            # Even though `.env` can't be overwritten
            assert_value_source("from_dotenv", Store.DOTENV)

        # Fails if no stores are supported
        with unsupported(Store.DOTENV), unsupported(Store.MELTANO_YML), unsupported(
            Store.DB,
        ):
            with pytest.raises(StoreNotSupportedError):
                set_value("nowhere")

            assert_value_source("from_dotenv", Store.DOTENV)

        # Falls back on system database when project is readonly
        monkeypatch.setattr(project, "readonly", True)
        metadata = set_value("from_db")
        assert metadata["store"] == Store.DB

        # Even though `.env` can't be overwritten
        assert_value_source("from_dotenv", Store.DOTENV)

    def test_unset(
        self,
        subject,
        project,
        unsupported,
        set_value_store,
        assert_value_source,
        monkeypatch,
        environment,
    ) -> None:
        set_value_store("from_dotenv", Store.DOTENV, name="password")

        set_value_store("from_dotenv", Store.DOTENV)
        set_value_store("from_meltano_yml", Store.MELTANO_YML)
        set_value_store("from_db", Store.DB)

        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            set_value_store("from_meltano_environment", Store.MELTANO_ENV)

        metadata = subject.unset("regular", ["regular"])
        assert metadata["store"] == Store.DB
        assert_value_source("from_default", Store.DEFAULT)

        assert_value_source("from_dotenv", Store.DOTENV, name="password")

        # Fails when store is not supported
        with unsupported(Store.DOTENV):
            with pytest.raises(StoreNotSupportedError):
                subject.unset("password", ["password"])

            assert_value_source("from_dotenv", Store.DOTENV, name="password")

        # Unsets even when there is technically no setting with that exact full name
        subject.manager_for(Store.MELTANO_YML).set(
            "custom.nested",
            ["custom", "nested"],
            "from_meltano_yml",
        )
        assert_value_source(None, Store.DEFAULT, name="custom")
        assert_value_source("from_meltano_yml", Store.MELTANO_YML, name="custom.nested")

        subject.unset("custom", ["custom"])
        assert_value_source(None, Store.DEFAULT, name="custom")
        assert_value_source(None, Store.DEFAULT, name="custom.nested")

        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            subject.manager_for(Store.MELTANO_ENV).set(
                "custom_in_env",
                ["custom_in_env"],
                "from_meltano_environment",
            )
            assert_value_source(
                "from_meltano_environment",
                Store.MELTANO_ENV,
                name="custom_in_env",
            )

        subject.unset("custom_in_env", ["custom_in_env"])
        assert_value_source(None, Store.DEFAULT, name="custom_in_env")

    def test_reset(
        self,
        subject,
        unsupported,
        set_value_store,
        assert_value_source,
        project,
        environment,
        monkeypatch,
    ) -> None:
        set_value_store("from_db", Store.DB)
        set_value_store("from_meltano_yml", Store.MELTANO_YML, name="unknown")
        set_value_store("from_dotenv", Store.DOTENV, name="password")
        set_value_store("from_db", Store.DB, name="env_specific")
        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            set_value_store(
                "from_meltano_environment",
                Store.MELTANO_ENV,
                name="dataops",
            )

        subject.reset()

        assert_value_source("from_default", Store.DEFAULT)
        assert_value_source(None, Store.DEFAULT, name="unknown")
        assert_value_source(None, Store.DEFAULT, name="password")
        assert_value_source(None, Store.DEFAULT, name="env_specific")
        assert_value_source(None, Store.DEFAULT, name="dataops")
        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            assert_value_source(
                "from_meltano_environment",
                Store.MELTANO_ENV,
                name="dataops",
            )

        # Fails silently when store is not supported
        set_value_store("from_dotenv", Store.DOTENV, name="password")
        with unsupported(Store.DOTENV):
            subject.reset()

            assert_value_source("from_dotenv", Store.DOTENV, name="password")

        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            set_value_store(
                "from_meltano_environment",
                Store.MELTANO_ENV,
                name="dataops",
            )
            assert_value_source(
                "from_meltano_environment",
                Store.MELTANO_ENV,
                name="dataops",
            )

        # Resetting without an active environment has no effect on config 'dataops'
        subject.reset()
        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            assert_value_source(
                "from_meltano_environment",
                Store.MELTANO_ENV,
                name="dataops",
            )

        # Resetting with an active environment works fine :)
        with monkeypatch.context() as mpc:
            mpc.setattr(project, "environment", environment)
            subject.reset()
            assert_value_source(None, Store.DEFAULT, name="dataops")


class TestMeltanoYmlStoreManager:
    @pytest.fixture()
    def subject(self, dummy_settings_service):
        manager = MeltanoYmlStoreManager(dummy_settings_service)
        yield manager
        manager.reset()

    def test_get(self, subject) -> None:
        def get():
            return subject.get(
                "regular",
                setting_def=subject.settings_service.find_setting("regular"),
            )

        assert get() == (None, {})

        subject.flat_config["basic"] = "alias_value"

        assert get() == ("alias_value", {"expandable": True, "key": "basic"})

        subject.flat_config.pop("basic")
        subject.flat_config["regular"] = "value"

        assert get() == ("value", {"expandable": True, "key": "regular"})

    def test_set(self, subject) -> None:
        def set_value(key, value):
            return subject.set(
                key,
                [key],
                value,
                setting_def=subject.settings_service.find_setting(key),
            )

        subject.flat_config["basic"] = "alias_value"

        set_value("basic", "new_alias_value")

        assert "regular" not in subject.flat_config
        assert subject.flat_config["basic"] == "new_alias_value"

        set_value("regular", "new_value")

        assert "basic" not in subject.flat_config
        assert subject.flat_config["regular"] == "new_value"

    def test_unset(self, subject) -> None:
        def unset_value(key):
            return subject.unset(
                key,
                [key],
                setting_def=subject.settings_service.find_setting(key),
            )

        def set_values() -> None:
            subject.flat_config["regular"] = "value"
            subject.flat_config["basic"] = "alias_value"

        set_values()
        unset_value("regular")

        assert "regular" not in subject.flat_config
        assert "basic" not in subject.flat_config

        set_values()
        unset_value("basic")

        assert "regular" not in subject.flat_config
        assert "basic" not in subject.flat_config


class TestMeltanoEnvironmentStoreManager(TestMeltanoYmlStoreManager):
    @pytest.fixture()
    def subject(self, dummy_settings_service, project):
        project.refresh(environment=Environment("testing", {}))
        manager = MeltanoEnvStoreManager(dummy_settings_service)
        try:
            yield manager
        finally:
            project.deactivate_environment()
            manager.reset()


class TestInheritedStoreManager:
    @pytest.fixture()
    def subject(self, dummy_settings_service):
        return InheritedStoreManager(dummy_settings_service)

    def test_get(self, subject, project) -> None:
        def get(key="regular"):
            return subject.get(
                key,
                setting_def=subject.settings_service.find_setting(key),
            )

        with pytest.raises(StoreNotSupportedError):
            get()

        inherited_settings = DummySettingsService(project)
        subject.settings_service._inherited_settings = inherited_settings

        # Default values are not inherited
        assert inherited_settings.get_with_source("regular") == (
            "from_default",
            Store.DEFAULT,
        )
        assert get() == (None, {})

        # Non-default values are inherited
        inherited_settings.set("regular", "$YML_VALUE", store=Store.MELTANO_YML)
        value, metadata = get()

        # Env vars are not expanded in values
        assert value == "$YML_VALUE"
        assert metadata["inherited_source"] is Store.MELTANO_YML
        # Env var expandability is inherited
        assert metadata["expandable"]

        inherited_settings.set("regular", "$DOTENV_VALUE", store=Store.DOTENV)
        value, metadata = get()
        assert value == "$DOTENV_VALUE"
        assert metadata["inherited_source"] is Store.DOTENV
        # Lack of env var expandability is inherited
        assert not metadata["expandable"]


class TestEnvStoreManager:
    @pytest.fixture()
    def subject(self, dummy_settings_service) -> EnvStoreManager:
        return EnvStoreManager(dummy_settings_service)

    def test_ensure_supported(self, subject: EnvStoreManager) -> None:
        with pytest.raises(StoreNotSupportedError, match="Store is not writable"):
            subject.ensure_supported(method="set")

        subject.ensure_supported(method="get")


class TestDotEnvStoreManager:
    @pytest.fixture()
    def subject(self, dummy_settings_service) -> DotEnvStoreManager:
        return DotEnvStoreManager(dummy_settings_service)

    def test_set_undefined_setting_failure(self, subject: DotEnvStoreManager) -> None:
        with pytest.raises(StoreNotSupportedError, match="Unknown setting"):
            subject.set("undefined", [], "my-value", setting_def=None)

    def test_unset_undefined_setting_failure(self, subject: DotEnvStoreManager) -> None:
        with pytest.raises(StoreNotSupportedError, match="Unknown setting"):
            subject.unset("undefined", [], setting_def=None)

    def test_reset_readonly_project_failure(
        self,
        project,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        monkeypatch.setattr(project, "readonly", True)
        settings_service = DummySettingsService(project)
        manager = DotEnvStoreManager(settings_service)

        with pytest.raises(
            StoreNotSupportedError,
            match="This Meltano project is deployed as read-only",
        ):
            manager.reset()



"""Various utils and formatters for log rendering control."""

from __future__ import annotations

import typing as t

import click
import structlog
from rich.console import Console
from rich.traceback import Traceback, install

from meltano.core.utils import get_no_color_flag

if t.TYPE_CHECKING:
    from types import TracebackType

    from structlog.types import Processor

install(suppress=[click])

TIMESTAMPER = structlog.processors.TimeStamper(fmt="iso")

LEVELED_TIMESTAMPED_PRE_CHAIN = (
    # Add the log level and a timestamp to the event_dict if the log entry
    # is not from structlog.
    structlog.stdlib.add_log_level,
    TIMESTAMPER,
)


def rich_exception_formatter_factory(
    color_system: t.Literal["auto", "standard", "256", "truecolor", "windows"] = "auto",
    *,
    no_color: bool | None = None,
    show_locals: bool = False,
) -> t.Callable[[t.TextIO, structlog.types.ExcInfo], None]:
    """Create an exception formatter for logging using the rich package.

    Examples:
    >>> rich_traceback = rich_exception_formatter_factory(color_system="truecolor")
    >>> plane_rich_traceback = rich_exception_formatter_factory(no_color=True)

    Args:
        color_system: The color system supported by your terminal.
        no_color: Enabled no color mode, or None to auto detect. Defaults to None.
        show_locals: Whether to show local variables in the traceback.

    Returns:
        Exception formatter function.
    """

    def _traceback(
        sio,  # noqa: ANN001
        exc_info: tuple[type[t.Any], BaseException, TracebackType | None],
    ) -> None:
        sio.write("\n")
        Console(file=sio, color_system=color_system, no_color=no_color).print(
            Traceback.from_exception(
                *exc_info,
                show_locals=show_locals,
            ),
        )

    return _traceback


def _process_formatter(processor: Processor) -> structlog.stdlib.ProcessorFormatter:
    """Use _process_formatter to configure a structlog.stdlib.ProcessFormatter.

    It will automatically add log level and timestamp fields to any log entries
    not originating from structlog.

    Args:
        processor: A structlog message processor such as
            `structlog.dev.ConsoleRenderer`.

    Returns:
        A configured log processor.
    """
    return structlog.stdlib.ProcessorFormatter(
        processor=processor,
        foreign_pre_chain=LEVELED_TIMESTAMPED_PRE_CHAIN,
    )


def console_log_formatter(
    *,
    colors: bool = False,
    show_locals: bool = False,
) -> structlog.stdlib.ProcessorFormatter:
    """Create a logging formatter for console rendering that supports colorization.

    Args:
        colors: Add color to output.
        show_locals: Whether to show local variables in the traceback.

    Returns:
        A configured console log formatter.
    """
    colors = colors and not get_no_color_flag()

    if colors:
        exception_formatter = rich_exception_formatter_factory(
            color_system="truecolor",
            show_locals=show_locals,
        )
    else:
        exception_formatter = rich_exception_formatter_factory(
            no_color=True,
            show_locals=show_locals,
        )

    return _process_formatter(
        structlog.dev.ConsoleRenderer(
            colors=colors,
            exception_formatter=exception_formatter,
        ),
    )


def key_value_formatter(
    *,
    sort_keys: bool = False,
    key_order: t.Sequence[str] | None = None,
    drop_missing: bool = False,
) -> structlog.stdlib.ProcessorFormatter:
    """Create a logging formatter that renders lines in key=value format.

    Args:
        sort_keys: Whether to sort keys when formatting.
        key_order: List of keys that should be rendered in this exact order.
            Missing keys will be rendered as None, extra keys depending on
            *sort_keys* and the dict class.
        drop_missing: When True, extra keys in *key_order* will be dropped
            rather than rendered as None.

    Returns:
        A configured key=value formatter.
    """
    return _process_formatter(
        processor=structlog.processors.KeyValueRenderer(
            sort_keys=sort_keys,
            key_order=key_order,
            drop_missing=drop_missing,
        ),
    )


def json_formatter() -> structlog.stdlib.ProcessorFormatter:
    """Create a logging formatter that renders lines in JSON format.

    Returns:
        A configured JSON formatter.
    """
    return _process_formatter(processor=structlog.processors.JSONRenderer())



"""YAML instance."""

from __future__ import annotations

import typing as t
from dataclasses import dataclass
from pathlib import Path

from ruamel.yaml import YAML, CommentedMap

from meltano.core.behavior.canonical import Canonical
from meltano.core.plugin import PluginType
from meltano.core.setting_definition import SettingKind
from meltano.core.utils import hash_sha256

if t.TYPE_CHECKING:
    import os

yaml = YAML()
yaml.default_flow_style = False

yaml.register_class(Canonical)
yaml.register_class(PluginType)
yaml.register_class(SettingKind)


@dataclass
class CachedCommentedMap:
    """The hash of the raw bytes of a YAML file, and its parsed content."""

    sha256: str
    data: CommentedMap


cache: dict[os.PathLike, CachedCommentedMap] = {}


def load(path: os.PathLike) -> CommentedMap:
    """Load the specified YAML file with caching.

    The cache is used if both the file path and its content hash match what is stored.

    Parameters:
        path: The path to the YAML file.

    Returns:
        The loaded YAML file.
    """
    path = Path(path).resolve()
    with path.open() as yaml_file:
        hashed = hash_sha256(yaml_file.read())

        if path in cache and cache[path].sha256 == hashed:
            return cache[path].data

        yaml_file.seek(0)
        parsed = yaml.load(yaml_file)

    cache[path] = CachedCommentedMap(hashed, parsed)
    return parsed


# Alias to provide a clean interface when using this
# module via `from meltano.core import yaml`
dump = yaml.dump



"""Manage Python virtual environments."""

from __future__ import annotations

import asyncio
import contextlib
import hashlib
import os
import platform
import shlex
import shutil
import subprocess
import sys
import typing as t
from asyncio.subprocess import Process
from functools import cached_property, lru_cache
from numbers import Number
from pathlib import Path

import structlog

from meltano.core.error import AsyncSubprocessError, MeltanoError

if t.TYPE_CHECKING:
    from collections.abc import Iterable

    from meltano.core.project import Project

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias  # noqa: ICN003

if sys.version_info < (3, 12):
    from typing_extensions import override
else:
    from typing import override  # noqa: ICN003


logger = structlog.stdlib.get_logger(__name__)

StdErrExtractor: TypeAlias = t.Callable[[Process], t.Awaitable[t.Union[str, None]]]


@lru_cache(maxsize=None)
def find_uv() -> str:
    """Find the `uv` executable.

    Tries to import the `uv` package and use its `find_uv_bin` function to find the
    `uv` executable. If that fails, falls back to using the `uv` executable on the
    system PATH. If it can't be found on the PATH, returns `"uv"`.

    Adapted from https://github.com/wntrblm/nox/blob/55c7eaf2eb03feb4a4b79e74966c542b75d61401/nox/virtualenv.py#L42-L54.

    Copyright 2016 Alethea Katherine Flowers

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Returns:
        A string representing the path to the `uv` executable.

    Raises:
        MeltanoError: The `uv` executable could not be found.
    """
    logger.warning("The uv backend is experimental")

    with contextlib.suppress(ImportError, FileNotFoundError):
        from uv import find_uv_bin

        return find_uv_bin()

    # Fall back to PATH.
    uv = shutil.which("uv")

    if not uv:
        error = "Could not find the 'uv' executable"
        instruction = "Please install 'meltano[uv]' or install 'uv' globally."
        raise MeltanoError(error, instruction)

    return uv


def _resolve_python_path(python: Path | str | None) -> str:
    python_path: str | None = None

    if python is None:
        python_path = sys.executable
    elif isinstance(python, Path):
        python_path = str(python.resolve())
    elif isinstance(python, Number):
        raise MeltanoError(
            "Python must be specified as an executable name or path, "  # noqa: EM102
            f"not the number {python!r}",
        )
    else:
        python_path = python if os.path.exists(python) else shutil.which(python)  # noqa: PTH110

    if python_path is None:
        raise MeltanoError(f"Python executable {python!r} was not found")  # noqa: EM102

    if not os.access(python_path, os.X_OK):
        raise MeltanoError(f"{python_path!r} is not executable")  # noqa: EM102

    return python_path


class VirtualEnv:
    """Info about a single virtual environment."""

    _SUPPORTED_PLATFORMS: t.ClassVar[set[str]] = {
        "Linux",
        "Darwin",
        "Windows",
    }

    def __init__(
        self,
        root: Path,
        python: Path | str | None = None,
    ):
        """Initialize the `VirtualEnv` instance.

        Args:
            root: The root directory of the virtual environment.
            python: The path to the Python executable to use, or name to find on the
                $PATH. Defaults to the Python executable running Meltano.

        Raises:
            MeltanoError: The current system is not supported.
        """
        self._system = platform.system()
        if self._system not in self._SUPPORTED_PLATFORMS:
            raise MeltanoError(f"Platform {self._system!r} not supported.")  # noqa: EM102
        self.root = root.resolve()
        self.python_path = _resolve_python_path(python)
        self.plugin_fingerprint_path = self.root / ".meltano_plugin_fingerprint"

    @cached_property
    def lib_dir(self) -> Path:
        """Return the lib directory of the virtual environment.

        Raises:
            MeltanoError: The current system is not supported.

        Returns:
            The lib directory of the virtual environment.
        """
        if self._system in {"Linux", "Darwin"}:
            return self.root / "lib"
        if self._system == "Windows":
            return self.root / "Lib"
        raise MeltanoError(f"Platform {self._system!r} not supported.")  # noqa: EM102

    @cached_property
    def bin_dir(self) -> Path:
        """Return the bin directory of the virtual environment.

        Raises:
            MeltanoError: The current system is not supported.

        Returns:
            The bin directory of the virtual environment.
        """
        if self._system in {"Linux", "Darwin"}:
            return self.root / "bin"
        if self._system == "Windows":
            return self.root / "Scripts"
        raise MeltanoError(f"Platform {self._system!r} not supported.")  # noqa: EM102

    @cached_property
    def site_packages_dir(self) -> Path:
        """Return the site-packages directory of the virtual environment.

        Raises:
            MeltanoError: The current system is not supported.

        Returns:
            The site-packages directory of the virtual environment.
        """
        if self._system in {"Linux", "Darwin"}:
            return (
                self.lib_dir
                / f"python{'.'.join(str(x) for x in self.python_version_tuple[:2])}"
                / "site-packages"
            )
        if self._system == "Windows":
            return self.lib_dir / "site-packages"
        raise MeltanoError(f"Platform {self._system!r} not supported.")  # noqa: EM102

    @cached_property
    def python_version_tuple(self) -> tuple[int, int, int]:
        """Return the Python version tuple of the virtual environment.

        Returns:
            The Python version tuple of the virtual environment.
        """
        if self.python_path == sys.executable:
            return sys.version_info[:3]
        return t.cast(
            tuple[int, int, int],
            tuple(
                int(x)
                for x in subprocess.run(
                    (
                        self.python_path,
                        "-c",
                        "import sys; print(*sys.version_info[:3])",
                    ),
                    stdout=subprocess.PIPE,
                ).stdout.split(b" ")
            ),
        )

    def read_fingerprint(self) -> str | None:
        """Get the fingerprint of the existing virtual environment.

        Returns:
            The fingerprint of the existing virtual environment if it exists.
            `None` otherwise.
        """
        if not self.plugin_fingerprint_path.exists():
            return None
        return self.plugin_fingerprint_path.read_text()

    def write_fingerprint(self, pip_install_args: t.Sequence[str]) -> None:
        """Save the fingerprint for this installation.

        Args:
            pip_install_args: The arguments being passed to `pip install`.
        """
        self.plugin_fingerprint_path.write_text(fingerprint(pip_install_args))


async def _extract_stderr(_) -> None:
    return None  # pragma: no cover


async def exec_async(*args, extract_stderr=_extract_stderr, **kwargs) -> Process:  # noqa: ANN001, ANN002, ANN003
    """Run an executable asynchronously in a subprocess.

    Args:
        args: Positional arguments for `asyncio.create_subprocess_exec`.
        extract_stderr: Async function that is provided the completed failed process,
            and returns its error string or `None`.
        kwargs: Keyword arguments for `asyncio.create_subprocess_exec`.

    Raises:
        AsyncSubprocessError: The command failed.

    Returns:
        The subprocess.
    """
    run = await asyncio.create_subprocess_exec(
        *args,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        **kwargs,
    )
    await run.wait()

    if run.returncode != 0:
        raise AsyncSubprocessError(
            "Command failed",  # noqa: EM101
            process=run,
            stderr=await extract_stderr(run),
        )

    return run


def fingerprint(pip_install_args: Iterable[str]) -> str:
    """Generate a hash identifying pip install args.

    Arguments are sorted and deduplicated before the hash is generated.

    Args:
        pip_install_args: Arguments for `pip install`.

    Returns:
        The SHA256 hash hex digest of the sorted set of pip install args.
    """
    return hashlib.sha256(" ".join(sorted(set(pip_install_args))).encode()).hexdigest()


class VenvService:
    """Manages virtual environments.

    The methods in this class are not thread-safe.
    """

    def __init__(
        self,
        *,
        project: Project,
        python: Path | str | None = None,
        namespace: str = "",
        name: str = "",
    ):
        """Initialize the `VenvService`.

        Args:
            project: The Meltano project.
            python: The path to the Python executable to use, or name to find on $PATH.
                Defaults to the Python executable running Meltano.
            namespace: The namespace for the venv, e.g. a Plugin type.
            name: The name of the venv, e.g. a Plugin name.
        """
        self.project = project
        self.namespace = namespace
        self.name = name
        self.venv = VirtualEnv(
            self.project.venvs_dir(namespace, name, make_dirs=False),
            python or project.settings.get("python"),
        )
        self.pip_log_path = self.project.logs_dir(
            "pip",
            self.namespace,
            self.name,
            "pip.log",
        ).resolve()

    async def install(
        self,
        pip_install_args: t.Sequence[str],
        *,
        clean: bool = False,
        env: dict[str, str | None] | None = None,
    ) -> None:
        """Configure a virtual environment, then run pip install with the given args.

        Args:
            pip_install_args: Arguments passed to `pip install`.
            clean: Whether to not attempt to use an existing virtual environment.
            env: Environment variables to pass to the subprocess.
        """
        if not clean and self.requires_clean_install(pip_install_args):
            logger.debug(
                f"Packages for '{self.namespace}/{self.name}' have changed so "  # noqa: G004
                "performing a clean install.",
            )
            clean = True

        self.clean_run_files()

        await self._pip_install(pip_install_args=pip_install_args, clean=clean, env=env)
        self.venv.write_fingerprint(pip_install_args)

    def requires_clean_install(self, pip_install_args: t.Sequence[str]) -> bool:
        """Determine whether a clean install is needed.

        Args:
            pip_install_args: The arguments being passed to `pip install`, used
                for fingerprinting the installation.

        Returns:
            Whether virtual environment doesn't exist or can't be reused.
        """

        # A generator is used to perform the checks lazily
        def checks():  # noqa: ANN202
            # The Python installation used to create this venv no longer exists
            yield not self.exec_path("python").exists()
            # The fingerprint of the venv does not match the pip install args
            existing_fingerprint = self.venv.read_fingerprint()
            yield existing_fingerprint is None
            yield existing_fingerprint != fingerprint(pip_install_args)

        return any(checks())

    def clean_run_files(self) -> None:
        """Destroy cached configuration files, if they exist."""
        run_dir = self.project.run_dir(self.name, make_dirs=False)

        try:
            for path in run_dir.iterdir():
                if path.is_dir():
                    shutil.rmtree(path)
                else:
                    path.unlink()
        except FileNotFoundError:
            logger.debug("No cached configuration files to remove")

    def clean(self) -> None:
        """Destroy the virtual environment, if it exists."""
        try:
            shutil.rmtree(self.venv.root)
            logger.debug(
                "Removed old virtual environment for '%s/%s'",
                self.namespace,
                self.name,
            )
        except FileNotFoundError:
            # If the VirtualEnv has never been created before do nothing
            logger.debug("No old virtual environment to remove")

    async def create_venv(
        self,
        *,
        extract_stderr: StdErrExtractor,
    ) -> Process:
        """Create a new virtual environment.

        Args:
            extract_stderr: Async function that is provided the completed failed
                process, and returns its error string or `None`.

        Returns:
            The Python process creating the virtual environment.
        """
        return await exec_async(
            sys.executable,
            "-m",
            "virtualenv",
            "--python",
            self.venv.python_path,
            str(self.venv.root),
            extract_stderr=extract_stderr,
        )

    async def create(self) -> Process:
        """Create a new virtual environment.

        Raises:
            AsyncSubprocessError: The virtual environment could not be created.

        Returns:
            The Python process creating the virtual environment.
        """
        logger.debug(f"Creating virtual environment for '{self.namespace}/{self.name}'")  # noqa: G004

        async def extract_stderr(proc: Process):  # noqa: ANN202
            return (await t.cast(asyncio.StreamReader, proc.stdout).read()).decode(
                "utf-8",
                errors="replace",
            )

        try:
            return await self.create_venv(extract_stderr=extract_stderr)
        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                f"Could not create the virtualenv for '{self.namespace}/{self.name}'",  # noqa: EM102
                process=err.process,
                stderr=await err.stderr,
            ) from err

    async def upgrade_installer(
        self,
        *,
        env: dict[str, str | None] | None = None,
    ) -> Process | None:
        """Upgrade the `pip` package to the latest version in the virtual environment.

        Args:
            env: Environment variables to pass to the subprocess.

        Raises:
            AsyncSubprocessError: Failed to upgrade pip to the latest version.

        Returns:
            The process running `pip install --upgrade ...`.
        """
        logger.debug(f"Upgrading pip for '{self.namespace}/{self.name}'")  # noqa: G004
        try:
            return await self._pip_install(("--upgrade", "pip"), env=env)
        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                "Failed to upgrade pip to the latest version.",  # noqa: EM101
                err.process,
            ) from err

    def exec_path(self, executable: str) -> Path:
        """Return the absolute path for the given executable in the virtual environment.

        Args:
            executable: The path to the executable relative to the venv bin directory.

        Returns:
            The venv bin directory joined to the provided executable.
        """
        absolute_executable = self.venv.bin_dir / executable
        if platform.system() != "Windows":
            return absolute_executable

        # On Windows, try using the '.exe' suffixed version if it exists. Use the
        # regular executable path as a fallback (and for backwards compatibility).
        absolute_executable_windows = absolute_executable.with_suffix(".exe")
        return (
            absolute_executable_windows
            if absolute_executable_windows.exists()
            else absolute_executable
        )

    async def install_pip_args(
        self,
        pip_install_args: t.Sequence[str],
        *,
        extract_stderr: StdErrExtractor = _extract_stderr,
        env: dict[str, str | None] | None = None,
    ) -> Process:
        """Return the `pip install` arguments to use.

        Args:
            pip_install_args: The arguments to pass to `pip install`.
            extract_stderr: Async function that is provided the completed failed
                process, and returns its error string or `None`.
            env: Environment variables to pass to the subprocess.

        Returns:
            The process running `pip install` with the provided args.
        """
        return await exec_async(
            str(self.exec_path("python")),
            "-m",
            "pip",
            "install",
            "--log",
            str(self.pip_log_path),
            *pip_install_args,
            extract_stderr=extract_stderr,
            env=env,
        )

    async def uninstall_package(self, package: str) -> Process:
        """Uninstall a single package.

        Args:
            package: The package name.

        Returns:
            The process running `pip uninstall` with the provided package.
        """
        return await exec_async(
            str(self.exec_path("python")),
            "-m",
            "pip",
            "uninstall",
            "--yes",
            package,
            extract_stderr=_extract_stderr,
        )

    async def handle_installation_error(
        self,
        err: AsyncSubprocessError,
    ) -> AsyncSubprocessError:
        """Handle an error that occurred during installation.

        Args:
            err: The error that occurred during installation.

        Returns:
            The error that occurred during installation with additional context.
        """
        logger.info(
            "Logged pip install output to %s",
            self.pip_log_path,
        )
        return AsyncSubprocessError(
            f"Failed to install plugin '{self.name}'.",
            err.process,
            stderr=await err.stderr,
        )

    async def _pip_install(
        self,
        pip_install_args: t.Sequence[str],
        *,
        clean: bool = False,
        env: dict[str, str | None] | None = None,
    ) -> Process:
        """Install a package using `pip` in the proper virtual environment.

        Args:
            pip_install_args: The arguments to pass to `pip install`.
            clean: Whether the installation should be done in a clean venv.
            env: Environment variables to pass to the subprocess.

        Raises:
            AsyncSubprocessError: The command failed.

        Returns:
            The process running `pip install` with the provided args.
        """
        if clean:
            self.clean()
            await self.create()
            await self.upgrade_installer(env=env)

        pip_install_args_str = shlex.join(pip_install_args)
        log_msg_prefix = (
            f"Upgrading with args {pip_install_args_str!r} in existing"
            if "--upgrade" in pip_install_args
            else f"Installing with args {pip_install_args_str!r} into"
        )
        logger.debug(
            f"{log_msg_prefix} virtual environment for '{self.namespace}/{self.name}'",  # noqa: G004
        )

        async def extract_stderr(proc: Process) -> str | None:  # pragma: no cover
            if not proc.stdout:
                return None

            return (await proc.stdout.read()).decode("utf-8", errors="replace")

        try:
            return await self.install_pip_args(
                pip_install_args,
                extract_stderr=extract_stderr,
                env=env,
            )
        except AsyncSubprocessError as err:
            raise await self.handle_installation_error(err) from err


class UvVenvService(VenvService):
    """Manages virtual environments using `uv`."""

    def __init__(self, *args: t.Any, **kwargs: t.Any):
        """Initialize the `UvVenvService`.

        Args:
            args: Positional arguments for the VenvService.
            kwargs: Keyword arguments for the VenvService.
        """
        super().__init__(*args, **kwargs)
        self.uv = find_uv()
        logger.debug("Using uv executable at %s", self.uv)

    @override
    async def upgrade_installer(
        self,
        *,
        env: dict[str, str | None] | None = None,
    ) -> Process | None:
        """No-op for `uv` virtual environments.

        Args:
            env: Environment variables to pass to the subprocess.
        """

    @override
    async def install_pip_args(
        self,
        pip_install_args: t.Sequence[str],
        *,
        extract_stderr: StdErrExtractor = _extract_stderr,
        env: dict[str, str | None] | None = None,
    ) -> Process:
        """Run `pip install` in the plugin's virtual environment.

        Args:
            pip_install_args: The arguments to pass to `pip install`.
            extract_stderr: Async function that is provided the completed failed
                process, and returns its error string or `None`.
            env: Environment variables to pass to the subprocess.

        Returns:
            The process running `pip install` with the provided args.
        """
        return await exec_async(
            self.uv,
            "pip",
            "install",
            "--python",
            str(self.exec_path("python")),
            *pip_install_args,
            extract_stderr=extract_stderr,
            env=env,
        )

    @override
    async def uninstall_package(self, package: str) -> Process:
        """Uninstall a single package using `uv`.

        Args:
            package: The package name.

        Returns:
            The process running `pip uninstall` with the provided package.
        """
        return await exec_async(
            self.uv,
            "pip",
            "uninstall",
            "--python",
            str(self.exec_path("python")),
            package,
        )

    @override
    async def handle_installation_error(
        self,
        err: AsyncSubprocessError,
    ) -> AsyncSubprocessError:
        """Handle an error that occurred during installation.

        Args:
            err: The subprocess error that occurred during installation.

        Returns:
            The error that occurred during installation with additional context.
        """
        return AsyncSubprocessError(
            f"Failed to install plugin '{self.name}'.",
            err.process,
            stderr=await err.stderr,
        )

    @override
    async def create_venv(
        self,
        *,
        extract_stderr: StdErrExtractor,
    ) -> Process:
        """Create a new virtual environment using `uv`.

        Args:
            extract_stderr: Async function that is provided the completed failed
                process, and returns its error string or `None`.

        Returns:
            The Python process creating the virtual environment.
        """
        return await exec_async(
            self.uv,
            "virtualenv",
            "--python",
            self.venv.python_path,
            str(self.venv.root),
            extract_stderr=extract_stderr,
        )



"""Manager for state that persists across runs of a BlockSet.

'State' in this module refers to _Singer_ state, which is held in a Job's
'payload' field. This is not to be confused with the Job's 'state' field,
which refers to a given job run's status, e.g. 'RUNNING' or 'FAILED'.
"""

from __future__ import annotations

import datetime
import json
import typing as t

import structlog

from meltano.core.job import Job, Payload, State
from meltano.core.job_state import SINGER_STATE_KEY, JobState
from meltano.core.project import Project
from meltano.core.state_store import state_store_manager_from_project_settings

if t.TYPE_CHECKING:
    from sqlalchemy.orm import Session

logger = structlog.getLogger(__name__)


class InvalidJobStateError(Exception):
    """Invalid job state is parsed."""


class StateService:
    """Meltano Service used to manage job state.

    Currently only manages Singer state for Extract and Load jobs.
    """

    def __init__(self, project: Project | None = None, session: Session | None = None):
        """Create a StateService object.

        Args:
            project: current meltano Project
            session: the session to use, if using SYSTEMDB state backend
        """
        self.project = project or Project.find()
        self.session = session
        self._state_store_manager = None

    def list_state(self, state_id_pattern: str | None = None):  # noqa: ANN201
        """List all state found in the db.

        Args:
            state_id_pattern: An optional glob-style pattern of state_ids to search for

        Returns:
            A dict with state_ids as keys and state payloads as values.
        """
        return {
            state_id: self.get_state(state_id)
            for state_id in self.state_store_manager.get_state_ids(state_id_pattern)
        }

    def _get_or_create_job(self, job: Job | str) -> Job:
        """If Job is passed, return it. If state_id is passed, create new and return.

        Args:
            job: either an existing Job to modify state for, or a state_id

        Raises:
            TypeError: if job is not of type Job or str

        Returns:
            A new job with given state_id, or the given Job
        """
        if isinstance(job, str):
            now = datetime.datetime.now(datetime.timezone.utc)
            return Job(
                job_name=job,
                state=State.STATE_EDIT,
                started_at=now,
                ended_at=now,
            )
        if isinstance(job, Job):
            return job
        raise TypeError("job must be of type Job or of type str")  # noqa: EM101

    @property
    def state_store_manager(self):  # noqa: ANN201
        """Initialize and return the correct StateStoreManager.

        Returns:
            StateStoreManager instance.
        """
        if not self._state_store_manager:
            self._state_store_manager = state_store_manager_from_project_settings(
                self.project.settings,
                self.session,
            )
        return self._state_store_manager

    @staticmethod
    def validate_state(state: dict[str, t.Any]) -> None:
        """Check that the given state str is valid.

        Args:
            state: the state to validate

        Raises:
            InvalidJobStateError: if supplied state is not valid singer state
        """
        if SINGER_STATE_KEY not in state:
            raise InvalidJobStateError(
                f"{SINGER_STATE_KEY} not found in top level of provided state",  # noqa: EM102
            )

    def add_state(
        self,
        job: Job | str,
        new_state: str,
        payload_flags: Payload = Payload.STATE,
        *,
        validate=True,  # noqa: ANN001
    ) -> None:
        """Add state for the given Job.

        Args:
            job: either an existing Job or a state_id that future runs may look
                up state for.
            new_state: the state to add for the given job.
            payload_flags: the payload_flags to set for the job
            validate: whether to validate the supplied state
        """
        new_state_dict = json.loads(new_state)
        if validate:
            self.validate_state(new_state_dict)
        state_to_add_to = self._get_or_create_job(job)
        state_to_add_to.payload = json.loads(new_state)
        state_to_add_to.payload_flags = payload_flags
        state_to_add_to.save(self.session)
        logger.debug(
            f"Added to state {state_to_add_to.job_name} state payload {new_state_dict}",  # noqa: G004
        )
        partial_state = (
            new_state_dict if payload_flags == Payload.INCOMPLETE_STATE else {}
        )
        completed_state = new_state_dict if payload_flags == Payload.STATE else {}
        job_state = JobState(
            state_id=state_to_add_to.job_name,
            partial_state=partial_state,
            completed_state=completed_state,
        )
        self.state_store_manager.set(job_state)

    def get_state(self, state_id: str):  # noqa: ANN201
        """Get state for the given state_id.

        Args:
            state_id: The state_id to get state for

        Returns:
            Dict representing state that would be used in the next run.
        """
        if state := self.state_store_manager.get(state_id=state_id):
            return json.loads(state.json_merged())
        return {}

    def set_state(
        self,
        state_id: str,
        new_state: str,
        *,
        validate: bool = True,
    ) -> None:
        """Set the state for the state_id.

        Args:
            state_id: the state_id to set state for
            new_state: the state to update to
            validate: Whether to validate the supplied state.
        """
        self.add_state(
            state_id,
            new_state,
            payload_flags=Payload.STATE,
            validate=validate,
        )

    def clear_state(self, state_id, *, save: bool = True) -> None:  # noqa: ANN001, ARG002
        """Clear the state for the state_id.

        Args:
            state_id: the state_id to clear state for
            save: Whether to immediately save the state
        """
        self.state_store_manager.clear(state_id)

    def merge_state(self, state_id_src: str, state_id_dst: str) -> None:
        """Merge state from state_id_src into state_id_dst.

        Args:
            state_id_src: the state_id to get state from
            state_id_dst: the state_id_to merge state onto
        """
        self.add_state(
            state_id_dst,
            json.dumps(self.get_state(state_id_src)),
            payload_flags=Payload.INCOMPLETE_STATE,
        )

    def copy_state(self, state_id_src: str, state_id_dst: str) -> None:
        """Copy state from Job state_id_src onto Job state_id_dst.

        Args:
            state_id_src: the state_id to get state from
            state_id_dst: the state_id_to copy state onto
        """
        self.set_state(state_id_dst, json.dumps(self.get_state(state_id_src)))

    def move_state(self, state_id_src: str, state_id_dst: str) -> None:
        """Move state from Job state_id_src to Job state_id_dst.

        Args:
            state_id_src: the state_id to get state from and clear
            state_id_dst: the state_id_to move state onto
        """
        self.set_state(state_id_dst, json.dumps(self.get_state(state_id_src)))
        self.clear_state(state_id_src)



from __future__ import annotations  # noqa: D104

from pathlib import Path

import meltano.core.sqlalchemy as types

MIGRATION_DIR = Path(__path__[0])
LOCK_PATH = MIGRATION_DIR.joinpath("db.lock")


# Exposing the types here enable us to create versioned
# version for migrations purpose. For instance, if one
# of these types change in `meltano.core`, we would then
# create a version for this type for all migrations that
# would follow.
GUID = types.GUID
JSONEncodedDict = types.JSONEncodedDict
IntFlag = types.IntFlag



"""Dialect-specific SQLAlchemy types."""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.mssql import DATETIME2
from sqlalchemy.engine import Connection  # noqa: TCH002


def get_dialect_name() -> str:
    """Get the dialect name from the Alembic op.

    Returns:
        The dialect name.
    """
    database_connection: Connection = op.get_bind()
    return database_connection.dialect.name


def datetime_for_dialect(dialect_name: str) -> DATETIME2 | sa.DateTime:
    """Get the datetime type for the given dialect.

    Args:
        dialect_name: The SQL dialect name.

    Returns:
        The datetime type.
    """
    # We need to use the DATETIME2 type for MSSQL, because the
    # default DATETIME type does not go back to the year 1.
    return DATETIME2 if dialect_name == "mssql" else sa.DateTime


def max_string_length_for_dialect(dialect_name: str) -> int | None:
    """Get the maximum string length for the given dialect.

    We need to limit the size of the string to avoid MySQL throwing an error.

    Args:
        dialect_name: The SQL dialect name.

    Returns:
        The maximum string length or None if there is no limit.
    """
    return 1024 if dialect_name == "mysql" else None



from __future__ import annotations

from pathlib import Path

import pytest


@pytest.mark.meta()
def test_pushd(tmp_path: Path, pushd):
    tmp_path.joinpath("a").mkdir(parents=True)
    tmp_path.joinpath("a", "b").mkdir(parents=True)

    pushd(tmp_path / "a")
    assert Path.cwd() == tmp_path / "a"

    popd = pushd("b")
    assert Path.cwd() == tmp_path / "a" / "b"

    popd()
    assert Path.cwd() == tmp_path / "a"



"""Class for invoking plugin validations."""

from __future__ import annotations

import typing as t
from abc import ABCMeta, abstractmethod
from enum import Enum

from meltano.core.plugin import PluginType
from meltano.core.plugin_invoker import PluginInvoker, invoker_factory

if t.TYPE_CHECKING:
    from sqlalchemy.orm.session import Session

    from meltano.core.project import Project

EXIT_CODE_OK = 0

T = t.TypeVar("T", bound="ValidationsRunner")


class ValidationOutcome(str, Enum):
    """Data validation outcome options."""

    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"

    @property
    def color(self) -> str:
        """Return terminal color for this outcome.

        Returns:
            The string name of a color for this outcome.
        """
        return "green" if self == self.SUCCESS else "red"

    @classmethod
    def from_exit_code(cls, exit_code: int):  # noqa: ANN206
        """Create validation outcome from an exit code.

        Args:
            exit_code: Exit code of this of this outcome.

        Returns:
            A string name of this outcome ("SUCCESS or "FAILURE")
        """
        return cls.SUCCESS if exit_code == EXIT_CODE_OK else cls.FAILURE


class ValidationsRunner(metaclass=ABCMeta):
    """Class to collect all validations defined for a plugin."""

    def __init__(
        self,
        invoker: PluginInvoker,
        tests_selection: dict[str, bool],
    ) -> None:
        """Create a validators runner for a plugin.

        Args:
            invoker: PluginInvoker to be used with this ValidationsRunner.
            tests_selection: Dict of selected tests.
        """
        self.invoker = invoker
        self.tests_selection = tests_selection

    @property
    def plugin_name(self) -> str:
        """Get underlying plugin name.

        Returns:
            The name of the plugin to run.
        """
        return self.invoker.plugin.name

    def select_test(self, name: str) -> None:
        """Mark a single test as selected.

        Args:
            name: Test (command) name.

        Raises:
            KeyError: If plugin test is not defined.
        """
        try:
            self.tests_selection[name] = True
        except KeyError as exc:
            raise KeyError(
                f"Plugin {self.plugin_name} does not have a test named '{name}'",  # noqa: EM102
            ) from exc

    def select_all(self) -> None:
        """Mark all plugin validators as selected."""
        for name in self.tests_selection:
            self.tests_selection[name] = True

    async def run_all(self, session: Session) -> dict[str, int]:
        """Run all validators defined in a plugin.

        Args:
            session: A SQLAlchemy ORM session.

        Returns:
            A mapping of validator names to exit codes.
        """
        if not self.tests_selection:
            return {}

        results = {}  # type: ignore[var-annotated]
        async with self.invoker.prepared(session):
            for name, selected in self.tests_selection.items():
                if selected:
                    results[name] = await self.run_test(name)
        return results

    @classmethod
    def collect(
        cls: type[T],
        project: Project,
        *,
        select_all: bool = True,
    ) -> dict[str, T]:
        """Collect all tests for CLI invocation.

        Args:
            project: A Meltano project object.
            select_all: Flag to select all validations by default.

        Returns:
            A mapping of plugin names to validation runners.
        """
        return {
            plugin.name: cls(
                invoker=invoker_factory(project, plugin),
                tests_selection={
                    test_name: select_all for test_name in plugin.test_commands
                },
            )
            for plugin in project.plugins.plugins()
            if plugin.type is not PluginType.FILES
        }

    @abstractmethod
    async def run_test(self, name: str) -> t.NoReturn:
        """Run a test command.

        Args:
            name: Test name.
        """
        raise NotImplementedError



"""State backends."""

from __future__ import annotations

import platform
import typing as t
from enum import Enum
from urllib.parse import urlparse

from meltano.core.db import project_engine
from meltano.core.state_store.azure import AZStorageStateStoreManager
from meltano.core.state_store.db import DBStateStoreManager
from meltano.core.state_store.filesystem import (
    LocalFilesystemStateStoreManager,
    WindowsFilesystemStateStoreManager,
)
from meltano.core.state_store.google import GCSStateStoreManager
from meltano.core.state_store.s3 import S3StateStoreManager

if t.TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from meltano.core.project_settings_service import ProjectSettingsService
    from meltano.core.state_store.base import StateStoreManager


class StateBackend(str, Enum):
    """State backend."""

    SYSTEMDB = "systemdb"
    # These strings should match the expected scheme for URIs of
    # the given type. E.g., filesystem state backends have a
    # file://<path>/<to>/<state directory> URI
    LOCAL_FILESYSTEM = "file"
    AZURE = "azure"
    S3 = "s3"
    GCS = "gs"

    @classmethod
    def backends(cls) -> list[StateBackend]:
        """List available state backends.

        Returns:
            List of available state backends.
        """
        return list(cls)

    @property
    def _managers(
        self,
    ) -> t.Mapping[str, type[StateStoreManager]]:
        """Get mapping of StateBackend to associated StateStoreManager.

        Returns:
            Mapping of StateBackend to associated StateStoreManager.
        """
        return {
            self.SYSTEMDB: DBStateStoreManager,
            self.LOCAL_FILESYSTEM: LocalFilesystemStateStoreManager,
            self.S3: S3StateStoreManager,
            self.AZURE: AZStorageStateStoreManager,
            self.GCS: GCSStateStoreManager,
        }

    @property
    def manager(
        self,
    ) -> type[StateStoreManager]:
        """Get the StateStoreManager associated with this StateBackend.

        Returns:
            The StateStoreManager associated with this StateBackend.
        """
        return self._managers[self]


def state_store_manager_from_project_settings(
    settings_service: ProjectSettingsService,
    session: Session | None = None,
) -> StateStoreManager:
    """Return a StateStoreManager based on the project's settings.

    Args:
        settings_service: the settings service to use
        session: the session to use if using default systemdb state backend

    Returns:
        The relevant StateStoreManager instance.
    """
    state_backend_uri: str = settings_service.get("state_backend.uri")
    parsed = urlparse(state_backend_uri)
    if state_backend_uri == StateBackend.SYSTEMDB:
        return DBStateStoreManager(
            session=session or project_engine(settings_service.project)[1](),
        )
    scheme = parsed.scheme
    # Get backend-specific settings
    # AND top-level state_backend settings
    setting_defs = filter(
        lambda setting_def: setting_def.name.startswith(
            f"state_backend.{'gcs' if scheme == 'gs' else scheme}",
        )
        or (
            setting_def.name.startswith("state_backend")
            and len(setting_def.name.split(".")) == 2
        ),
        settings_service.setting_definitions,
    )
    settings = (setting_def.name for setting_def in setting_defs)
    backend = StateBackend(scheme).manager
    if scheme == StateBackend.LOCAL_FILESYSTEM and "Windows" in platform.system():
        backend = WindowsFilesystemStateStoreManager
    kwargs = {name.split(".")[-1]: settings_service.get(name) for name in settings}
    return backend(**kwargs)



"""Locked plugin definition service."""

from __future__ import annotations

import typing as t

from meltano.core.plugin import BasePlugin, PluginDefinition, PluginRef, PluginType
from meltano.core.plugin.base import StandalonePlugin
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.factory import base_plugin_factory
from meltano.core.plugin_repository import PluginRepository

if t.TYPE_CHECKING:
    from meltano.core.project import Project


class LockedDefinitionService(PluginRepository):
    """PluginRepository implementation for local files."""

    def __init__(self, project: Project) -> None:
        """Initialize the service.

        Args:
            project: The Meltano project.
        """
        self.project = project

    def find_definition(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant_name: str | None = None,
    ) -> PluginDefinition:
        """Find a locked plugin definition.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant_name: The plugin variant name.

        Returns:
            The plugin definition.

        Raises:
            PluginNotFoundError: If the plugin definition could not be found.
        """
        path = self.project.plugin_lock_path(
            plugin_type,
            plugin_name,
            variant_name=variant_name,
        )
        try:
            standalone = StandalonePlugin.parse_json_file(path)
        except FileNotFoundError as err:
            raise PluginNotFoundError(PluginRef(plugin_type, plugin_name)) from err
        return PluginDefinition.from_standalone(standalone)

    def find_base_plugin(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant: str | None = None,
    ) -> BasePlugin:
        """Get the base plugin for a project plugin.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant: The plugin variant.

        Returns:
            The base plugin.
        """
        plugin = self.find_definition(
            plugin_type,
            plugin_name,
            variant_name=variant,
        )

        return base_plugin_factory(plugin, plugin.variants[0])

    def get_plugins_of_type(
        self,
        plugin_type: PluginType,
    ) -> t.Iterable[PluginDefinition]:
        """Get all plugin definitions of a given type.

        Args:
            plugin_type: The plugin type.

        Yields:
            Locked plugin definitions for the given type.
        """
        plugin_type_dir = self.project.root_plugins_dir(plugin_type.value)
        yield from (
            PluginDefinition.from_standalone(StandalonePlugin.parse_json_file(lockfile))
            for lockfile in plugin_type_dir.glob("*.lock")
        )



from __future__ import annotations

import json
import logging
import os
import tempfile
import typing as t
import uuid
from pathlib import Path

import mock
import pytest
from mock import AsyncMock

from meltano.core.block.blockset import BlockSetValidationError
from meltano.core.block.extract_load import (
    ELBContext,
    ELBContextBuilder,
    ExtractLoadBlocks,
    generate_state_id,
)
from meltano.core.block.ioblock import IOBlock
from meltano.core.block.singer import SingerBlock
from meltano.core.environment import Environment
from meltano.core.job import Job, Payload, State
from meltano.core.job_state import STATE_ID_COMPONENT_DELIMITER
from meltano.core.logging import OutputLogger
from meltano.core.plugin import PluginType
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.project_plugins_service import PluginAlreadyAddedException
from meltano.core.runner import RunnerError
from meltano.core.runner.singer import SingerRunner

if t.TYPE_CHECKING:
    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project

TEST_STATE_ID = "test_job"
MOCK_STATE_MESSAGE = json.dumps({"type": "STATE"})
MOCK_RECORD_MESSAGE = json.dumps({"type": "RECORD"})


def create_plugin_files(config_dir: Path, plugin: ProjectPlugin):
    for file in plugin.config_files.values():
        Path(os.path.join(config_dir, file)).touch()  # noqa: PTH118

    return config_dir


@pytest.fixture()
def test_job(session) -> Job:
    return Job(
        job_name=TEST_STATE_ID,
        state=State.SUCCESS,
        payload_flags=Payload.STATE,
        payload={"singer_state": {"bookmarks": []}},
    ).save(session)


@pytest.fixture()
def output_logger() -> OutputLogger:
    return OutputLogger("test.log")


@pytest.fixture()
def elb_context(project, session, test_job, output_logger) -> ELBContext:
    ctx = ELBContext(
        project=project,
        job=test_job,
        base_output_logger=output_logger,
    )
    ctx.session = session
    return ctx


class TestELBContext:
    def test_elt_run_dir_is_returned(
        self,
        project,
        test_job,
        elb_context: ELBContext,
    ) -> None:
        expected_path = project.job_dir(test_job.job_name, str(test_job.run_id))
        assert elb_context.elt_run_dir == Path(expected_path)


class TestELBContextBuilder:
    @pytest.fixture()
    def target_postgres(self, project_add_service):
        try:
            return project_add_service.add(PluginType.LOADERS, "target-postgres")
        except PluginAlreadyAddedException as err:
            return err.plugin

    @pytest.mark.usefixtures("target")
    def test_builder_returns_elb_context(self, project, session, tap) -> None:
        """Ensure that builder is returning ELBContext and not itself."""
        builder = ELBContextBuilder(project)
        builder.session = session

        assert isinstance(builder.context(), ELBContext)
        assert isinstance(builder.make_block(tap).invoker.context, ELBContext)

    def test_make_block_returns_valid_singer_block(
        self,
        project,
        session,
        tap,
        target,
    ) -> None:
        """Ensure that calling make_block returns a valid SingerBlock."""
        builder = ELBContextBuilder(project)
        builder.session = session

        block = builder.make_block(tap)
        assert block.string_id == tap.name
        assert block.producer
        assert not block.consumer

        block = builder.make_block(target)
        assert block.string_id == target.name
        assert block.consumer
        assert not block.producer

    def test_make_block_tracks_envs(self, project, session, tap, target) -> None:
        """Ensure that calling make_block correctly stacks env vars."""
        builder = ELBContextBuilder(project)
        builder.session = session

        block = builder.make_block(tap)
        assert block.string_id == tap.name
        initial_dict = builder._env.copy()

        block2 = builder.make_block(target)
        assert block2.string_id == target.name
        assert initial_dict.items() <= builder._env.items()
        assert builder._env.items() >= block.context.env.items()
        assert builder._env.items() >= block2.context.env.items()

    @pytest.mark.asyncio()
    async def test_validate_envs(self, project, session, tap, target_postgres) -> None:
        """Ensure that expected environment variables are present."""
        builder = ELBContextBuilder(project)
        builder.session = session

        block = builder.make_block(tap)
        assert block.string_id == tap.name
        assert block.producer
        assert not block.consumer

        async with block.invoker.prepared(session):
            tap_env = block.invoker.env()

        assert tap_env["MELTANO_EXTRACTOR_NAME"] == tap.name
        assert tap_env["MELTANO_EXTRACTOR_NAMESPACE"] == tap.namespace
        assert tap_env["MELTANO_EXTRACTOR_VARIANT"] == tap.variant

        block = builder.make_block(target_postgres)
        assert block.string_id == target_postgres.name
        assert block.consumer
        assert not block.producer

        async with block.invoker.prepared(session):
            target_env = block.invoker.env()

        assert target_env["MELTANO_LOADER_NAME"] == target_postgres.name
        assert target_env["MELTANO_LOADER_NAMESPACE"] == target_postgres.namespace
        assert target_env["MELTANO_LOADER_VARIANT"] == target_postgres.variant

        assert target_env["MELTANO_LOAD_HOST"] == os.getenv(
            "TARGET_POSTGRES_HOST",
            "localhost",
        )

        assert (
            target_env["MELTANO_LOAD_DEFAULT_TARGET_SCHEMA"]
            == target_env["MELTANO_EXTRACT__LOAD_SCHEMA"]
            == target_env["MELTANO_EXTRACTOR_NAMESPACE"]
        )


class TestExtractLoadBlocks:
    @pytest.fixture()
    def log_level_debug(self):
        root_logger = logging.getLogger()  # noqa: TID251
        log_level = root_logger.level
        try:
            root_logger.setLevel(logging.DEBUG)
            yield
        finally:
            root_logger.setLevel(log_level)

    @pytest.fixture()
    def log(self, tmp_path: Path):
        return tempfile.NamedTemporaryFile(mode="w+", dir=tmp_path)

    @pytest.fixture()
    def tap_config_dir(self, tmp_path: Path, tap) -> Path:
        create_plugin_files(tmp_path, tap)
        return tmp_path

    @pytest.fixture()
    def mapper_config_dir(self, tmp_path: Path, tap) -> Path:
        create_plugin_files(tmp_path, tap)
        return tmp_path

    @pytest.fixture()
    def target_config_dir(self, tmp_path: Path, target) -> Path:
        create_plugin_files(tmp_path, target)
        return tmp_path

    @pytest.fixture()
    def subject(self, session, elb_context):
        Job(
            job_name=TEST_STATE_ID,
            state=State.SUCCESS,
            payload_flags=Payload.STATE,
            payload={"singer_state": {"bookmarks": []}},
        ).save(session)

        return SingerRunner(elb_context)

    @pytest.fixture()
    def process_mock_factory(self):
        def _factory(name):
            process_mock = mock.Mock()
            process_mock.name = name
            process_mock.wait = AsyncMock(return_value=0)
            return process_mock

        return _factory

    @pytest.fixture()
    def tap_process(self, process_mock_factory, tap):
        tap = process_mock_factory(tap)
        tap.stdout.readline = AsyncMock(return_value="{}")
        tap.wait = AsyncMock(return_value=0)
        return tap

    @pytest.fixture()
    def mapper_process(self, process_mock_factory, mapper):
        mapper = process_mock_factory(mapper)
        mapper.stdout.readline = AsyncMock(return_value="{}")
        mapper.wait = AsyncMock(return_value=0)
        return mapper

    @pytest.fixture()
    def target_process(self, process_mock_factory, target):
        target = process_mock_factory(target)
        target.stdout.readline = AsyncMock(return_value="{}")
        target.wait = AsyncMock(return_value=0)
        return target

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("session", "subject", "log", "log_level_debug")
    async def test_link_io(
        self,
        tap_config_dir,
        target_config_dir,
        mapper_config_dir,
        tap,
        target,
        mapper,
        tap_process,
        target_process,
        mapper_process,
        plugin_invoker_factory,
        elb_context,
    ) -> None:
        tap_process.sterr.at_eof.side_effect = True
        tap_process.stdout.at_eof.side_effect = (False, False, True)
        tap_process.stdout.readline = AsyncMock(
            side_effect=(
                b"%b" % json.dumps({"key": "value"}).encode(),
                b"%b" % MOCK_RECORD_MESSAGE.encode(),
            ),
        )

        mapper_process.sterr.at_eof.side_effect = True
        mapper_process.stdout.at_eof.side_effect = (False, False, True)
        mapper_process.stdout.readline = AsyncMock(
            side_effect=(
                b"%b" % json.dumps({"key": "mapper-mocked-value"}).encode(),
                b"%b" % MOCK_RECORD_MESSAGE.encode(),
            ),
        )

        tap_invoker = plugin_invoker_factory(tap, config_dir=tap_config_dir)
        mapper_invoker = plugin_invoker_factory(mapper, config_dir=mapper_config_dir)
        target_invoker = plugin_invoker_factory(target, config_dir=target_config_dir)

        invoke_async = AsyncMock(
            side_effect=(tap_process, mapper_process, target_process),
        )
        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            blocks = (
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=tap_invoker,
                    plugin_args=[],
                ),
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=mapper_invoker,
                    plugin_args=[],
                ),
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=target_invoker,
                    plugin_args=[],
                ),
            )

            elb = ExtractLoadBlocks(elb_context, blocks)
            elb.validate_set()

            for block in elb.blocks:
                await block.pre(elb.context)
                await block.start()

            await elb._link_io()

            # Explicitly check the counts of each block's output to ensure they
            # are linked count is going to be logger + 1 for next blocks stdin
            assert len(elb.blocks[0].outputs) == 2

            # block0 should write output to block1 stdin
            assert elb.blocks[1].stdin in elb.blocks[0].outputs

            assert len(elb.blocks[2].outputs) == 1  # logger only
            # block1 should write output to block2 stdin
            assert elb.blocks[2].stdin in elb.blocks[1].outputs

            # block2 should write output to logger and no where else
            assert len(elb.blocks[2].outputs) == 1

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("session", "subject", "log")
    async def test_extract_load_block(
        self,
        tap_config_dir,
        target_config_dir,
        mapper_config_dir,
        tap,
        target,
        mapper,
        tap_process,
        target_process,
        mapper_process,
        plugin_invoker_factory,
        elb_context,
    ) -> None:
        tap_process.sterr.at_eof.side_effect = True
        tap_process.stdout.at_eof.side_effect = (False, False, True)
        tap_process.stdout.readline = AsyncMock(
            side_effect=(
                b"%b" % json.dumps({"key": "value"}).encode(),
                b"%b" % MOCK_RECORD_MESSAGE.encode(),
            ),
        )

        mapper_process.sterr.at_eof.side_effect = True
        mapper_process.stdout.at_eof.side_effect = (False, False, True)
        mapper_process.stdout.readline = AsyncMock(
            side_effect=(
                b"%b" % json.dumps({"key": "mapper-value"}).encode(),
                b"%b" % MOCK_RECORD_MESSAGE.encode(),
            ),
        )

        tap_invoker = plugin_invoker_factory(tap, config_dir=tap_config_dir)
        mapper_invoker = plugin_invoker_factory(mapper, config_dir=mapper_config_dir)
        target_invoker = plugin_invoker_factory(target, config_dir=target_config_dir)

        invoke_async = AsyncMock(
            side_effect=(tap_process, mapper_process, target_process),
        )
        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            blocks = (
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=tap_invoker,
                    plugin_args=[],
                ),
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=mapper_invoker,
                    plugin_args=[],
                ),
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=target_invoker,
                    plugin_args=[],
                ),
            )

            elb = ExtractLoadBlocks(elb_context, blocks)
            elb.validate_set()
            await elb.run()

            assert tap_process.wait.called
            assert tap_process.stdout.readline.called

            assert mapper_process.wait.called
            assert mapper_process.stdout.readline.called
            assert mapper_process.stdin.writeline.called

            assert target_process.wait.called
            assert target_process.stdin.writeline.called

            # sanity check to verify that we saw mapper output and not tap output
            first_write = target_process.stdin.writeline.call_args_list[0]
            assert "mapper" in first_write[0][0]

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("session", "subject")
    async def test_elb_validation(
        self,
        tap_config_dir,
        target_config_dir,
        tap,
        target,
        tap_process,
        target_process,
        plugin_invoker_factory,
        elb_context,
    ) -> None:
        tap_process.sterr.at_eof.side_effect = True
        tap_process.stdout.at_eof.side_effect = (False, False, True)
        tap_process.stdout.readline = AsyncMock(
            side_effect=(
                b"%b" % json.dumps({"key": "value"}).encode(),
                b"%b" % MOCK_RECORD_MESSAGE.encode(),
            ),
        )

        tap_invoker = plugin_invoker_factory(tap, config_dir=tap_config_dir)
        target_invoker = plugin_invoker_factory(target, config_dir=target_config_dir)

        invoke_async = AsyncMock(side_effect=(tap_process, target_process))
        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            blocks = (
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=tap_invoker,
                    plugin_args=[],
                ),
            )

            elb = ExtractLoadBlocks(elb_context, blocks)

            with pytest.raises(
                BlockSetValidationError,
                match="^.*: last block in set should not be a producer",
            ):
                elb.validate_set()

            blocks = (
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=target_invoker,
                    plugin_args=[],
                ),
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=tap_invoker,
                    plugin_args=[],
                ),
            )
            elb = ExtractLoadBlocks(elb_context, blocks)
            with pytest.raises(
                BlockSetValidationError,
                match="^.*: first block in set should not be consumer",
            ):
                elb.validate_set()

            blocks = (
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=tap_invoker,
                    plugin_args=[],
                ),
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=tap_invoker,
                    plugin_args=[],
                ),
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=target_invoker,
                    plugin_args=[],
                ),
            )
            elb = ExtractLoadBlocks(elb_context, blocks)
            with pytest.raises(
                BlockSetValidationError,
                match="^.*: intermediate blocks must be producers AND consumers",
            ):
                elb.validate_set()

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("session", "subject")
    async def test_elb_with_job_context(
        self,
        tap_config_dir,
        mapper_config_dir,
        target_config_dir,
        tap,
        mapper,
        target,
        tap_process,
        mapper_process,
        target_process,
        plugin_invoker_factory,
        elb_context,
        project,
    ) -> None:
        tap_process.sterr.at_eof.side_effect = True
        tap_process.stdout.at_eof.side_effect = (False, False, True)
        tap_process.stdout.readline = AsyncMock(
            side_effect=(
                b"%b" % json.dumps({"key": "value"}).encode(),
                b"%b" % MOCK_RECORD_MESSAGE.encode(),
            ),
        )

        mapper_process.sterr.at_eof.side_effect = True
        mapper_process.stdout.at_eof.side_effect = (False, False, True)
        mapper_process.stdout.readline = AsyncMock(
            side_effect=(
                b"%b" % json.dumps({"key": "mapper-value"}).encode(),
                b"%b" % MOCK_RECORD_MESSAGE.encode(),
            ),
        )

        tap_invoker = plugin_invoker_factory(tap, config_dir=tap_config_dir)
        mapper_invoker = plugin_invoker_factory(mapper, config_dir=mapper_config_dir)
        target_invoker = plugin_invoker_factory(target, config_dir=target_config_dir)

        project.refresh(environment=Environment(name="test"))

        invoke_async = AsyncMock(
            side_effect=(tap_process, mapper_process, target_process),
        )
        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            blocks = (
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=tap_invoker,
                    plugin_args=[],
                ),
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=mapper_invoker,
                    plugin_args=[],
                ),
                SingerBlock(
                    block_ctx=elb_context,
                    project=elb_context.project,
                    plugin_invoker=target_invoker,
                    plugin_args=[],
                ),
            )

            elb = ExtractLoadBlocks(elb_context, blocks)
            elb.validate_set()

            assert elb.context.job.job_name == "test:tap-mock-to-target-mock"

            # Just to be sure, we double-check that state_id is the same for each block
            for block in blocks:
                assert block.context.job.job_name == "test:tap-mock-to-target-mock"

            elb.run_with_job = AsyncMock()

            await elb.run()
            assert elb.run_with_job.call_count == 1

    def test_custom_run_id(
        self,
        tap,
        target,
        project_with_environment: Project,
    ) -> None:
        run_id = uuid.UUID("12345678-1234-5678-1234-567812345678")
        builder = ELBContextBuilder(project_with_environment).with_run_id(run_id)
        assert builder._run_id == run_id

        elb_context = builder.context()
        tap_block = builder.make_block(tap)
        target_block = builder.make_block(target)
        extract_load_blocks = ExtractLoadBlocks(elb_context, (tap_block, target_block))
        assert extract_load_blocks.context.update_state
        assert extract_load_blocks.context.job is not None
        assert extract_load_blocks.context.job.run_id == run_id


class TestExtractLoadUtils:
    def test_generate_state_id(self) -> None:
        block1 = mock.Mock(spec=IOBlock)
        block1.string_id = "block1"

        block2 = mock.Mock(spec=IOBlock)
        block2.string_id = "block2"

        project = mock.Mock()
        project.environment = Environment(name="test")

        # Verify that a state ID is generated correctly given an active
        # environment and optional suffix.
        assert (
            generate_state_id(project, None, block1, block2) == "test:block1-to-block2"
        )
        assert (
            generate_state_id(project, "suffix", block1, block2)
            == "test:block1-to-block2:suffix"
        )

    def test_generate_state_id_no_environment(self) -> None:
        block1 = mock.Mock(spec=IOBlock)
        block1.string_id = "block1"

        block2 = mock.Mock(spec=IOBlock)
        block2.string_id = "block2"

        project = mock.Mock()
        project.environment = None

        # Verify an error is raised when attempting to generate a state ID
        # with no active environment.
        with pytest.raises(RunnerError):
            generate_state_id(project, None, block1, block2)

    def test_generate_state_id_component_contains_delimiter(self) -> None:
        block1 = mock.Mock(spec=IOBlock)
        block1.string_id = "block1"

        block2 = mock.Mock(spec=IOBlock)
        block2.string_id = "block2"

        project = mock.Mock()
        project.environment = Environment(name="test")

        # Verify an error is raised when attempting to generate a state ID
        # with a component that contains the defined delimiter string.
        with pytest.raises(RunnerError):
            generate_state_id(project, STATE_ID_COMPONENT_DELIMITER, block1, block2)



"""Migration and system db management."""

from __future__ import annotations

import logging
import typing as t

import click
import structlog
from alembic import command
from alembic.config import Config
from alembic.runtime.migration import MigrationContext
from alembic.script import ScriptDirectory

from meltano.migrations import LOCK_PATH, MIGRATION_DIR

if t.TYPE_CHECKING:
    from sqlalchemy.engine import Engine

logger = structlog.stdlib.get_logger(__name__)
SPLAT = "*"


class MigrationError(Exception):
    """Generic class for migration errors."""


class MigrationUneededException(Exception):
    """No migrations are needed."""


class MigrationService:
    """Migration service."""

    def __init__(self, engine: Engine) -> None:
        """Initialize the migration service.

        Args:
            engine: The sqlalchemy engine to use for the migration and checks.
        """
        self.engine = engine

    def ensure_migration_needed(
        self,
        script: ScriptDirectory,
        context: MigrationContext,
        target_revision: str,
    ) -> None:
        """Ensure that a migration of the system database is actually needed.

        Args:
            script: The alembic script directory.
            context: The migration context.
            target_revision: The desired target revision.

        Raises:
            MigrationUneededException: If no migration is needed.
        """
        current_head = context.get_current_revision()

        for rev in script.iterate_revisions(current_head, "base"):
            if rev.revision == target_revision:
                raise MigrationUneededException

    def upgrade(  # too many expression and too complex
        self,
        *,
        silent: bool = False,
    ) -> None:
        """Upgrade to the latest revision.

        Args:
            silent: If true, don't print anything.

        Raises:
            MigrationError: If the upgrade fails.
        """
        with self.engine.begin() as conn:
            cfg = Config()

            # this connection is used in `env.py` for the migrations
            cfg.attributes["connection"] = conn
            cfg.set_main_option("script_location", str(MIGRATION_DIR))
            script = ScriptDirectory.from_config(cfg)
            # let's make sure we actually need to migrate

            migration_logger = logging.getLogger("alembic.runtime.migration")  # noqa: TID251
            original_log_level = migration_logger.getEffectiveLevel()
            if silent:
                migration_logger.setLevel(logging.ERROR)

            context = MigrationContext.configure(conn)

            try:
                # try to find the locked version
                head = LOCK_PATH.read_text().strip()
                self.ensure_migration_needed(script, context, head)

                if not silent:
                    click.secho(f"Upgrading database to {head}")
                command.upgrade(cfg, head)

                if silent:
                    migration_logger.setLevel(original_log_level)
            except FileNotFoundError as ex:
                raise MigrationError(
                    "Cannot upgrade the system database, revision lock not found.",  # noqa: EM101
                ) from ex
            except MigrationUneededException:
                if not silent:
                    click.secho("System database up-to-date.")
            except Exception as ex:
                logger.exception(str(ex))
                raise MigrationError(
                    "Cannot upgrade the system database. It might be corrupted or "  # noqa: EM101
                    "was created before database migrations where introduced (v0.34.0)",
                ) from ex



"""Container management."""

from __future__ import annotations

import asyncio
import signal
import typing as t

from structlog.stdlib import get_logger

if t.TYPE_CHECKING:
    from aiodocker.containers import DockerContainer

    from meltano.core.container.container_spec import ContainerSpec


logger = get_logger(__name__)


def stop_container(container: DockerContainer) -> None:
    """Stop a Docker container.

    Args:
        container: Running container.
    """
    logger.info("Stopping container", container_id=container.id)
    asyncio.ensure_future(container.stop())  # noqa: RUF006


class ContainerService:
    """Wrapper for container interaction."""

    async def run_container(
        self,
        spec: ContainerSpec,
        name: str,
        *,
        env: dict | None = None,
        pull: bool = False,
    ) -> dict:
        """Run a Docker container.

        Args:
            spec: Command container spec.
            name: Container name.
            env: Environment mapping for the container run.
            pull: Pull image from registry.

        Returns:
            Docker container information after execution.
        """
        import aiodocker

        async with aiodocker.Docker() as docker:
            if pull:
                await docker.images.pull(spec.image)

            config = spec.get_docker_config(additional_env=env)

            logger.info("Running command in container", container_name=name)
            container = await docker.containers.run(config, name=name)
            loop = asyncio.get_running_loop()
            loop.add_signal_handler(signal.SIGINT, lambda: stop_container(container))

            async for line in container.log(follow=True, stdout=True, stderr=True):
                logger.info(line.rstrip())

            try:
                await container.wait()
            except Exception as exc:
                logger.exception("Container run failed", exc_info=exc)
            finally:
                info = await container.show()
                loop.remove_signal_handler(signal.SIGINT)
                await container.delete(force=True)

        return info



"""Plugin invoker class."""

from __future__ import annotations

import asyncio
import enum
import os
import typing as t
import uuid
from contextlib import asynccontextmanager

from structlog.stdlib import get_logger

from meltano.core.container.container_service import ContainerService
from meltano.core.error import Error
from meltano.core.plugin.config_service import PluginConfigService
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.settings_service import FeatureFlags
from meltano.core.tracking import Tracker
from meltano.core.utils import EnvVarMissingBehavior, expand_env_vars
from meltano.core.venv_service import VenvService, VirtualEnv

if t.TYPE_CHECKING:
    from pathlib import Path

    from sqlalchemy.orm import Session

    from meltano.core.logging.utils import SubprocessOutputWriter
    from meltano.core.plugin import PluginRef
    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project

logger = get_logger(__name__)


def invoker_factory(project, plugin: ProjectPlugin, *args, **kwargs):  # noqa: ANN001, ANN002, ANN003, ANN201
    """Instantiate a plugin invoker from a project plugin.

    Args:
        project: Meltano project.
        plugin: Plugin instance.
        args: Invoker constructor positional arguments.
        kwargs: Invoker constructor keyword arguments.

    Returns:
        A plugin invoker.
    """
    cls = PluginInvoker

    if hasattr(plugin, "invoker_class"):
        cls = plugin.invoker_class

    return cls(project, plugin, *args, **kwargs)


class InvokerError(Error):
    """Generic plugin invoker error."""


class ExecutableNotFoundError(InvokerError):
    """The executable could not be found."""

    def __init__(self, plugin: PluginRef, executable: str):
        """Initialize ExecutableNotFoundError.

        Args:
            plugin: Meltano plugin reference.
            executable: Plugin command executable.
        """
        plugin_type_descriptor = plugin.type.descriptor.capitalize()
        plugin_type = plugin.type.singular
        super().__init__(
            f"Executable '{executable}' could not be found. "
            f"{plugin_type_descriptor} '{plugin.name}' may not have "
            "been installed yet using "
            f"`meltano install {plugin_type} {plugin.name}`, "
            "or the executable name may be incorrect.",
        )


class InvokerNotPreparedError(InvokerError):
    """Occurs when `invoke` is called before `prepare`."""


class UnknownCommandError(InvokerError):
    """Occurs when `invoke` is called in command mode with an undefined command."""

    def __init__(self, plugin: PluginRef, command):  # noqa: ANN001
        """Initialize UnknownCommandError.

        Args:
            plugin: Meltano plugin reference.
            command: Plugin command name.
        """
        self.plugin = plugin
        self.command = command

    def __str__(self) -> str:
        """Return error message.

        Returns:
            String representation of this exception.
        """
        if self.plugin.supported_commands:
            supported_commands = ", ".join(self.plugin.supported_commands)
            desc = f"supports the following commands: {supported_commands}"
        else:
            desc = "does not define any commands."
        plugin_type_descriptor = self.plugin.type.descriptor.capitalize()
        plugin_name = self.plugin.name
        return " ".join(
            [
                f"Command '{self.command}' could not be found.",
                f"{plugin_type_descriptor} '{plugin_name}'",
                desc,
            ],
        )


class PluginInvoker:
    """This class handles the invocation of a `ProjectPlugin` instance."""

    class StdioSource(str, enum.Enum):
        """Describes the available unix style std io sources."""

        STDIN = "stdin"
        STDOUT = "stdout"
        STDERR = "stderr"

    def __init__(
        self,
        project: Project,
        plugin: ProjectPlugin,
        context: t.Any | None = None,  # noqa: ANN401
        output_handlers: dict | None = None,
        run_dir: Path | None = None,
        config_dir: Path | None = None,
        plugin_config_service: PluginConfigService | None = None,
        plugin_settings_service: PluginSettingsService | None = None,
    ):
        """Create a new plugin invoker.

        Args:
            project: Meltano Project.
            plugin: Meltano Plugin.
            context: Invocation context.
            output_handlers: Logging and output handlers.
            run_dir: Execution directory.
            config_dir: Configuration files directory.
            plugin_config_service: Plugin Configuration manager.
            plugin_settings_service: Plugin Settings manager.
        """
        self.project = project
        self.tracker = Tracker(project)
        self.plugin = plugin
        self.context = context
        self.output_handlers = output_handlers

        if plugin.pip_url:
            self.venv_service = VenvService(
                project=project,
                python=self.plugin.python,
                name=plugin.venv_name,
                namespace=plugin.type,
            )
        else:
            self.venv_service = None

        self.plugin_config_service = plugin_config_service or PluginConfigService(
            plugin,
            config_dir or self.project.plugin_dir(plugin),
            run_dir or self.project.run_dir(plugin.name),
        )

        self.settings_service = plugin_settings_service or PluginSettingsService(
            project,
            plugin,
        )

        self._prepared = False
        self.plugin_config = {}
        self.plugin_config_processed = {}
        self.plugin_config_extras = {}
        self.plugin_config_env = {}

    @property
    def capabilities(self):  # noqa: ANN201
        """Get plugin immutable capabilities.

        Makes sure the capabilities are immutable from the `PluginInvoker` interface.

        Returns:
            The set of plugin capabilities.
        """
        return frozenset(self.plugin.capabilities)

    @property
    def files(self) -> dict[str, Path]:
        """Get all config and output files of the plugin.

        Returns:
            A mapping of file IDs to file names.
        """
        plugin_files = {**self.plugin.config_files, **self.plugin.output_files}
        return {
            _key: self.plugin_config_service.run_dir.joinpath(filename)
            for _key, filename in plugin_files.items()
        }

    async def prepare(self, session: Session) -> None:
        """Prepare plugin config.

        Args:
            session: Database session.
        """
        self.plugin_config = self.settings_service.as_dict(
            extras=False,
            session=session,
        )
        self.plugin_config_processed = self.settings_service.as_dict(
            extras=False,
            process=True,
            session=session,
        )
        self.plugin_config_extras = self.settings_service.as_dict(
            extras=True,
            session=session,
        )
        self.plugin_config_env = self.settings_service.as_env(session=session)
        async with self.plugin.trigger_hooks("configure", self, session):
            self.plugin_config_service.configure()
            self._prepared = True

    async def cleanup(self) -> None:
        """Reset the plugin config."""
        self.plugin_config = {}
        self.plugin_config_processed = {}
        self.plugin_config_extras = {}
        self.plugin_config_env = {}

        async with self.plugin.trigger_hooks("cleanup", self):
            self._prepared = False

    @asynccontextmanager
    async def prepared(self, session: Session):  # noqa: ANN201
        """Context manager that prepares plugin config.

        Args:
            session: Database session.

        Yields:
            Yields to the caller, then resetting the config.
        """
        try:
            await self.prepare(session)
            yield
        finally:
            await self.cleanup()

    def exec_path(self, executable: str | None = None) -> str | Path:
        """Return the absolute path to the executable.

        Uses the plugin executable if none is specified.

        Args:
            executable: Optional executable string.

        Returns:
            Full path to the executable.
        """
        executable = executable or self.plugin.executable
        if not self.venv_service:
            if "/" not in executable.replace("\\", "/"):
                # Expect executable on path
                return executable

            # Return executable relative to project directory
            return self.project.root.joinpath(executable)

        # Return executable within venv
        return self.venv_service.exec_path(executable)

    def exec_args(self, *args, command=None, env=None):  # noqa: ANN001, ANN002, ANN201
        """Materialize the arguments to be passed to the executable.

        Args:
            args: Optional plugin args.
            command: Plugin command name.
            env: Environment variables

        Returns:
            List of plugin invocation arguments.
        """
        env = env or {}
        executable = self.exec_path()
        if command:
            command_config = self.find_command(command)
            plugin_args = command_config.expanded_args(command, env)
            if command_config.executable:
                executable = self.exec_path(command_config.executable)
        else:
            plugin_args = self.plugin.exec_args(self)

        return [str(arg) for arg in (executable, *plugin_args, *args)]

    def find_command(self, name):  # noqa: ANN001, ANN201
        """Find a Command by name.

        Args:
            name: Command name.

        Returns:
            Command instance.

        Raises:
            UnknownCommandError: If command is not defined.
        """
        try:
            return self.plugin.all_commands[name]
        except KeyError as err:
            raise UnknownCommandError(self.plugin, name) from err

    def env(self):  # noqa: ANN201
        """Environment variable mapping.

        Returns:
            Dictionary of environment variables.
        """
        with self.project.settings.feature_flag(
            FeatureFlags.STRICT_ENV_VAR_MODE,
            raise_error=False,
        ) as strict_env_var_mode:
            # Expand root env w/ os.environ
            expanded_project_env = {
                **expand_env_vars(
                    self.project.settings.env,
                    os.environ,
                    if_missing=EnvVarMissingBehavior(
                        strict_env_var_mode,
                    ),
                ),
                **expand_env_vars(
                    self.settings_service.project.dotenv_env,
                    os.environ,
                    if_missing=EnvVarMissingBehavior(strict_env_var_mode),
                ),
            }
            # Expand active env w/ expanded root env
            expanded_active_env = (
                expand_env_vars(
                    self.settings_service.project.environment.env,
                    expanded_project_env,
                    if_missing=EnvVarMissingBehavior(strict_env_var_mode),
                )
                if self.settings_service.project.environment
                else {}
            )

            # Expand root plugin env w/ expanded active env
            expanded_root_plugin_env = expand_env_vars(
                self.settings_service.plugin.env,
                expanded_active_env,
                if_missing=EnvVarMissingBehavior(strict_env_var_mode),
            )

            # Expand active env plugin env w/ expanded root plugin env
            expanded_active_env_plugin_env = (
                expand_env_vars(
                    self.settings_service.environment_plugin_config.env,
                    expanded_root_plugin_env,
                    if_missing=EnvVarMissingBehavior(strict_env_var_mode),
                )
                if self.settings_service.environment_plugin_config
                else {}
            )

        env = {
            **expanded_project_env,
            **self.project.dotenv_env,
            **self.settings_service.env,
            **self.plugin_config_env,
            **expanded_root_plugin_env,
            **expanded_active_env,
            **expanded_active_env_plugin_env,
            **self.tracker.env,
        }

        # Ensure Meltano venv is not inherited
        env.pop("VIRTUAL_ENV", None)
        env.pop("PYTHONPATH", None)
        if self.venv_service:
            # Switch to plugin-specific venv
            venv = VirtualEnv(
                self.project.venvs_dir(
                    self.plugin.type,
                    self.plugin.name,
                    make_dirs=False,
                ),
            )
            venv_dir = str(venv.bin_dir)
            env["VIRTUAL_ENV"] = str(venv.root)
            env["PATH"] = os.pathsep.join([venv_dir, env["PATH"]])

        return env

    def Popen_options(self) -> dict[str, t.Any]:  # noqa: N802
        """Get options for subprocess.Popen.

        Returns:
            Mapping of subprocess options.
        """
        return {}

    @asynccontextmanager
    async def _invoke(
        self,
        *args: str,
        require_preparation: bool = True,
        env: dict[str, t.Any] | None = None,
        command: str | None = None,
        **kwargs,  # noqa: ANN003
    ) -> t.Generator[list[str], dict[str, t.Any], dict[str, t.Any]]:
        env = env or {}

        if require_preparation and not self._prepared:
            raise InvokerNotPreparedError

        async with self.plugin.trigger_hooks("invoke", self, args):
            popen_options = {**self.Popen_options(), **kwargs}
            popen_env = {**self.env(), **env}
            popen_args = self.exec_args(*args, command=command, env=popen_env)
            logger.debug(f"Invoking: {popen_args}")  # noqa: G004

            try:
                yield (popen_args, popen_options, popen_env)
            except FileNotFoundError as err:
                raise ExecutableNotFoundError(
                    self.plugin,
                    self.plugin.executable,
                ) from err

    async def invoke_async(self, *args, **kwargs) -> asyncio.subprocess.Process:  # noqa: ANN002, ANN003
        """Invoke a command.

        Args:
            args: Positional arguments.
            kwargs: Keyword arguments.

        Returns:
            Subprocess.
        """
        async with self._invoke(*args, **kwargs) as (
            popen_args,
            popen_options,
            popen_env,
        ):
            return await asyncio.create_subprocess_exec(
                *popen_args,
                **popen_options,
                env=popen_env,
            )

    async def invoke_docker(
        self,
        plugin_command: str,
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> int:
        """Invoke a containerized command.

        Args:
            plugin_command: Plugin command name.
            args: Command line invocation arguments.
            kwargs: Command line invocation keyword arguments.

        Raises:
            ValueError: If the command doesn't declare a container spec.

        Returns:
            The container run exit code.
        """
        command_config = self.find_command(plugin_command)

        if not command_config.container_spec:
            raise ValueError("Command is missing a container spec")  # noqa: EM101

        spec = command_config.container_spec
        service = ContainerService()

        logger.debug("Running containerized command", command=plugin_command)
        async with self._invoke(*args, **kwargs) as (_proc_args, _, proc_env):
            plugin_name = self.plugin.name
            random_id = uuid.uuid4()
            name = f"meltano-{plugin_name}--{plugin_command}-{random_id}"

            info = await service.run_container(spec, name, env=proc_env)

        return info["State"]["ExitCode"]

    async def dump(self, file_id: str) -> str:
        """Dump a plugin file by ID.

        Args:
            file_id: Dump this file identifier.

        Returns:
            File contents.

        Raises:
            __cause__: If file is not found.
        """
        try:
            if file_id != "config":
                async with self._invoke():
                    return self.files[file_id].read_text()

            return self.files[file_id].read_text()
        except ExecutableNotFoundError as err:  # . Allow "useless" except.
            # Unwrap FileNotFoundError
            raise err.__cause__ from None

    def add_output_handler(self, src: str, handler: SubprocessOutputWriter) -> None:
        """Append an output handler for a given stdio stream.

        Args:
            src: stdio source you'd like to subscribe, likely either 'stdout'
                or 'stderr'
            handler: A `StreamWriter` or object matching the
                `utils.SubprocessOutputWriter` protocol
        """
        if self.output_handlers:
            self.output_handlers[src].append(handler)
        else:
            self.output_handlers = {src: [handler]}



"""New Project Initialization CLI."""

from __future__ import annotations

from pathlib import Path

import click
import structlog

from meltano.cli.params import database_uri_option
from meltano.cli.utils import InstrumentedCmd
from meltano.core.project_init_service import ProjectInitService
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.tracking import Tracker
from meltano.core.tracking.contexts import CliContext, CliEvent

EXTRACTORS = "extractors"
LOADERS = "loaders"
ALL = "all"

logger = structlog.stdlib.get_logger(__name__)
path_type = click.Path(file_okay=False, path_type=Path)


@click.command(cls=InstrumentedCmd, short_help="Create a new Meltano project.")
@click.pass_context
@click.argument("project_directory", required=False, type=path_type)
@click.option(
    "--no-usage-stats",
    help="Do not send anonymous usage stats.",
    is_flag=True,
)
@click.option(
    "--force",
    help="Overwrite `meltano.yml` if it exists in the project directory.",
    is_flag=True,
)
@database_uri_option
def init(ctx, project_directory: Path, *, no_usage_stats: bool, force: bool) -> None:  # noqa: ANN001
    """Create a new Meltano project.

    \b
    Read more at https://docs.meltano.com/reference/command-line-interface#init

    """  # noqa: D301
    if not project_directory:
        click.echo("We need a project name to get started!")
        project_directory = click.prompt(
            "Enter a name now to create a Meltano project",
            type=path_type,
        )

    if ctx.obj["project"]:
        root = ctx.obj["project"].root
        logger.warning(f"Found meltano project at: {root}")  # noqa: G004

    if no_usage_stats:
        ProjectSettingsService.config_override["send_anonymous_usage_stats"] = False

    init_service = ProjectInitService(project_directory)

    project = init_service.init(force=force)
    init_service.echo_instructions(project)

    # since the project didn't exist, tracking was not initialized in cli.py
    # but now that we've created a project we can set up telemetry and fire events.
    tracker = Tracker(project)
    tracker.add_contexts(CliContext.from_click_context(ctx))
    tracker.track_command_event(CliEvent.started)
    ctx.obj["tracker"] = tracker



"""Plugin Lockfile Service."""

from __future__ import annotations

import json
import typing as t
from hashlib import sha256

from structlog.stdlib import get_logger

from meltano.core.plugin.base import PluginRef, StandalonePlugin

if t.TYPE_CHECKING:
    from pathlib import Path

    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project

logger = get_logger(__name__)


class LockfileAlreadyExistsError(Exception):
    """Raised when a plugin lockfile already exists."""

    def __init__(self, message: str, path: Path, plugin: PluginRef):
        """Create a new LockfileAlreadyExistsError.

        Args:
            message: The error message.
            path: The path to the existing lockfile.
            plugin: The plugin that was locked.
        """
        self.path = path
        self.plugin = plugin
        super().__init__(message)


class PluginLock:
    """Plugin lockfile."""

    def __init__(self, project: Project, plugin: ProjectPlugin) -> None:
        """Create a new PluginLock.

        Args:
            project: The project.
            plugin: The plugin to lock.
        """
        self.project = project
        self.definition = plugin.definition
        self.variant = self.definition.find_variant(plugin.variant)

        self.path = self.project.plugin_lock_path(
            self.definition.type,
            self.definition.name,
            variant_name=self.variant.name,
        )

    def save(self) -> None:
        """Save the plugin lockfile."""
        locked_def = StandalonePlugin.from_variant(self.variant, self.definition)

        with self.path.open("w") as lockfile:
            json.dump(locked_def.canonical(), lockfile, indent=2)

    def load(
        self,
        *,
        create: bool = False,
        loader: t.Callable = lambda x: StandalonePlugin(**json.load(x)),
    ) -> StandalonePlugin:
        """Load the plugin lockfile.

        Args:
            create: Create the lockfile if it does not yet exist.
            loader: Function to process the lock file. Defaults to constructing
                a `StandalonePlugin` instance.

        Raises:
            FileNotFoundError: The lock file was not found at its expected path.

        Returns:
            The loaded plugin.
        """

        def _load():  # noqa: ANN202
            with self.path.open() as lockfile:
                return loader(lockfile)

        try:
            return _load()
        except FileNotFoundError:
            if create:
                self.save()
                return _load()
            raise

    @property
    def sha256_checksum(self) -> str:
        """Get the checksum of the lockfile.

        Returns:
            The checksum of the lockfile.
        """
        return sha256(self.path.read_bytes()).hexdigest()


class PluginLockService:
    """Plugin Lockfile Service."""

    def __init__(self, project: Project):
        """Create a new Plugin Lockfile Service.

        Args:
            project: The Meltano project.
        """
        self.project = project

    def save(
        self,
        plugin: ProjectPlugin,
        *,
        exists_ok: bool = False,
    ) -> None:
        """Save the plugin lockfile.

        Args:
            plugin: The plugin definition to save.
            exists_ok: Whether raise an exception if the lockfile already exists.

        Raises:
            LockfileAlreadyExistsError: If the lockfile already exists and is not
                flagged for overwriting.
        """
        plugin_lock = PluginLock(self.project, plugin)

        if plugin_lock.path.exists() and not exists_ok:
            raise LockfileAlreadyExistsError(
                f"Lockfile already exists: {plugin_lock.path}",  # noqa: EM102
                plugin_lock.path,
                plugin,
            )

        plugin_lock.save()

        logger.debug("Locked plugin definition", path=plugin_lock.path)



from __future__ import annotations

import typing as t

import mock

from asserts import assert_cli_runner
from meltano.cli import cli

if t.TYPE_CHECKING:
    from collections import Counter

    from click.testing import CliRunner
    from requests_mock import Mocker as RequestsMocker

    from meltano.core.project import Project


class TestCliHub:
    def test_ping(
        self,
        project: Project,
        cli_runner: CliRunner,
        hub_request_counter: Counter,
    ) -> None:
        with mock.patch(
            "requests.adapters.HTTPAdapter.send",
            project.hub_service.session.get_adapter(
                project.hub_service.hub_api_url,
            ).send,
        ):
            result = cli_runner.invoke(cli, ("hub", "ping"))

        assert_cli_runner(result)

        assert hub_request_counter["/orchestrators/index"] == 1
        assert (
            f"Successfully connected to the Hub at {project.hub_service.hub_api_url!r}"
            in result.stdout
        )

    def test_ping_unreachable(
        self,
        project: Project,
        cli_runner: CliRunner,
        hub_request_counter: Counter,
        requests_mock: RequestsMocker,
    ) -> None:
        hub_api = project.hub_service.hub_api_url
        requests_mock.get(
            hub_api,
            exc=ConnectionError("Connection refused"),
        )
        result = cli_runner.invoke(cli, ("hub", "ping"))
        assert f"Error: Failed to connect to the Hub at {hub_api!r}" in result.stderr
        assert result.exit_code == 1
        assert not hub_request_counter
